# -*- coding: utf-8 -*-
"""
@author: Oleg Ovchinnikov
"""
from __future__ import division, print_function, absolute_import


def apply_select_channel(file_in_h5,img_num,channel_num):
    
    import numpy as np
    import h5py as h5

    main_h5_handle = h5.File(file_in_h5,'r+')
    image_path="/Frame_%04i/Channel_%02i" % (img_num,channel_num)
    image_path="%s/Raw_Data" % (image_path)
    
    h5_image=main_h5_handle.get(image_path)
    img2 = np.empty(h5_image.shape, dtype=h5_image.dtype)
    h5_image.read_direct(img2)
    
    sec_i_ref=h5_image.attrs.get("Spectroscopic_Indices");
    sec_i_reg=h5_image.attrs.get("Spectroscopic_Indices_Region");
    sec_v_ref=h5_image.attrs.get("Spectroscopic_Values");
    sec_v_reg=h5_image.attrs.get("Spectroscopic_Values_Region");
    pos_i_ref=h5_image.attrs.get("Position_Indices");
    pos_i_reg=h5_image.attrs.get("Position_Indices_Region");
    pos_v_ref=h5_image.attrs.get("Position_Values");
    pos_v_reg=h5_image.attrs.get("Position_Values_Region");
    current_ref=h5_image.ref
    current_reg=h5_image.regionref[0:len(img2)]
    
    spec_ind=main_h5_handle[sec_i_ref];
    spec_ind=spec_ind[sec_i_reg];
    
    posi_ind=main_h5_handle[pos_i_ref];
    posi_ind=posi_ind[pos_i_reg];
 
  
                
                        
    image_path="/Frame_%04i/Channel_Current/Filter_Step_0000" % (img_num)
    
    try:
        main_h5_handle.__delitem__(image_path)
    except:
        temp=1;
      
        
    image_path="%s/Filtered_Image" % (image_path)
    main_h5_handle[image_path]=img2;
    h5_image_new = main_h5_handle.get(image_path);
    h5_new_attrs = h5_image_new.attrs
    h5_new_attrs["Filter_Name"]="Original";
    h5_new_attrs["Number_Of_Variables"] = 0;


    h5_new_attrs["Spectroscopic_Indices"]=sec_i_ref;
    h5_new_attrs["Spectroscopic_Indices_Region"]=sec_i_reg;
    h5_new_attrs["Spectroscopic_Values"]=sec_v_ref;
    h5_new_attrs["Spectroscopic_Values_Region"]=sec_v_reg;
    h5_new_attrs["Position_Indices"]=pos_i_ref;
    h5_new_attrs["Position_Indices_Region"]=pos_i_reg;
    h5_new_attrs["Position_Values"]=pos_v_ref;
    h5_new_attrs["Position_Values_Region"]=pos_v_reg;
    h5_new_attrs["Parent"]=current_ref;
    h5_new_attrs["Parent_Region"]=current_reg;
    
    image_path="/Frame_%04i/Channel_Current" % (img_num)
    
    path_main=main_h5_handle.get(image_path)  
    path_attrs=path_main 
    
    channel_name="Channel_%02i" % (channel_num)
    path_main=main_h5_handle.get(image_path)  
    path_attrs=path_main.attrs
    path_attrs["Origin"]=current_ref;
    path_attrs["Origin_Region"]=current_reg;
    path_attrs["Origin_Name"]=channel_name;

    main_h5_handle.close()
    return 1

def apply_wiener_filter(file_in_h5,img_num,filter_num):
    
    import numpy as np
    import h5py as h5
    

       
    main_h5_handle = h5.File(file_in_h5,'r+')
    image_path="/Frame_%04i/Channel_Current" % (img_num)
    for x in range(0,filter_num+1): 
        image_path="%s/Filter_Step_%04i" % (image_path,x)        
    image_path="%s/Filtered_Image" % (image_path)
    
    h5_image=main_h5_handle.get(image_path)
    img2 = np.empty(h5_image.shape, dtype=h5_image.dtype)
    h5_image.read_direct(img2)
    
    sec_i_ref=h5_image.attrs.get("Spectroscopic_Indices");
    sec_i_reg=h5_image.attrs.get("Spectroscopic_Indices_Region");
    sec_v_ref=h5_image.attrs.get("Spectroscopic_Values");
    sec_v_reg=h5_image.attrs.get("Spectroscopic_Values_Region");
    pos_i_ref=h5_image.attrs.get("Position_Indices");
    pos_i_reg=h5_image.attrs.get("Position_Indices_Region");
    pos_v_ref=h5_image.attrs.get("Position_Values");
    pos_v_reg=h5_image.attrs.get("Position_Values_Region");
    current_ref=h5_image.ref
    current_reg=h5_image.regionref[0:len(img2)]
    
    spec_ind=main_h5_handle[sec_i_ref];
    spec_ind=spec_ind[sec_i_reg];
    
    posi_ind=main_h5_handle[pos_i_ref];
    posi_ind=posi_ind[pos_i_reg];
    img = np.empty([max(posi_ind[:,0]),max(posi_ind[:,1])],dtype=h5_image.dtype)
    img[posi_ind[:,0]-1,posi_ind[:,1]-1]=img2[0:len(img2),0]


    
    
    # test shape doe wiener filter
    s=1.1 # constant for now change latter 
    
    size=img.shape 
    h = np.empty(size)
    width=size[0]
    height=size[1]
    for x in range(0,width):
        for y in range(0,height):
            h[x,y]=(s/2*3.141592)**(-s*np.sqrt(x**2+y**2))
    h = h/ np.sum(np.sum(h))    
    H = np.fft.fft2(h)
    K = np.linspace(.001,1,100)
    G = np.fft.fft2(img)
    errorV=np.empty([1,100]) 
    errorV=errorV[0]
    for k1 in range(0,100):
        W = np.conj(H)/(np.abs(H)**2 + K[k1])
        F = W*G
        R = np.abs(np.fft.ifft2(F))
        errorV[k1]=np.std(np.reshape(R,[1,width*height]))
        
        
    # use best option 
    minv=np.min(errorV)
    minl=np.where(errorV==minv)
    minl=minl[0]
    minl=minl[0]
    W = np.conj(H)/(np.abs(H)**2 + K[minl])
    F = W*G
    img= np.fft.ifft2(F)
    
    img=img.reshape(len(img2),1)
    img=abs(img);
    img=np.real(img);
    
    image_path="/Frame_%04i/Channel_Current" % (img_num)
    for x in range(0,filter_num+2): 
        image_path="%s/Filter_Step_%04i" % (image_path,x)
    
    try:
        main_h5_handle.__delitem__(image_path)
    except:
        temp=1;
        
    image_path="%s/Filtered_Image" % (image_path)
    main_h5_handle[image_path]=img;
    h5_image_new = main_h5_handle.get(image_path);
    h5_new_attrs = h5_image_new.attrs
    h5_new_attrs["Filter_Name"]="Wiener";
    h5_new_attrs["Number_Of_Variables"]=0;
    
    
    h5_new_attrs["Spectroscopic_Indices"]=sec_i_ref;
    h5_new_attrs["Spectroscopic_Indices_Region"]=sec_i_reg;
    h5_new_attrs["Spectroscopic_Values"]=sec_v_ref;
    h5_new_attrs["Spectroscopic_Values_Region"]=sec_v_reg;
    h5_new_attrs["Position_Indices"]=pos_i_ref;
    h5_new_attrs["Position_Indices_Region"]=pos_i_reg;
    h5_new_attrs["Position_Values"]=pos_v_ref;
    h5_new_attrs["Position_Values_Region"]=pos_v_reg;
    h5_new_attrs["Parent"]=current_ref;
    h5_new_attrs["Parent_Region"]=current_reg;
    
    main_h5_handle.close()
    
    return 1

def apply_gaussian_corr_filter(file_in_h5,img_num,filter_num,gauss_width,gauss_box_width):
    
    import numpy as np
    import h5py as h5
    
       
    main_h5_handle = h5.File(file_in_h5,'r+')
    image_path="/Frame_%04i/Channel_Current" % (img_num)
    for x in range(0,filter_num+1): 
        image_path="%s/Filter_Step_%04i" % (image_path,x)        
    image_path="%s/Filtered_Image" % (image_path)
    
    h5_image=main_h5_handle.get(image_path)
    img2 = np.empty(h5_image.shape, dtype=h5_image.dtype)
    h5_image.read_direct(img2)
    
    sec_i_ref=h5_image.attrs.get("Spectroscopic_Indices");
    sec_i_reg=h5_image.attrs.get("Spectroscopic_Indices_Region");
    sec_v_ref=h5_image.attrs.get("Spectroscopic_Values");
    sec_v_reg=h5_image.attrs.get("Spectroscopic_Values_Region");
    pos_i_ref=h5_image.attrs.get("Position_Indices");
    pos_i_reg=h5_image.attrs.get("Position_Indices_Region");
    pos_v_ref=h5_image.attrs.get("Position_Values");
    pos_v_reg=h5_image.attrs.get("Position_Values_Region");
    current_ref=h5_image.ref
    current_reg=h5_image.regionref[0:len(img2)]
    
    spec_ind=main_h5_handle[sec_i_ref];
    spec_ind=spec_ind[sec_i_reg];
    
    posi_ind=main_h5_handle[pos_i_ref];
    posi_ind=posi_ind[pos_i_reg];
    img = np.empty([max(posi_ind[:,0]),max(posi_ind[:,1])],dtype=h5_image.dtype)
    img[posi_ind[:,0]-1,posi_ind[:,1]-1]=img2[0:len(img2),0]
    
    gauss_cell=[1,0,0,gauss_width,gauss_width,0];
    
    s1=len(img[:,0]);
    s2=len(img[0,:]);
    new_deconv=np.zeros([s1,s2]);
    
    for k1 in range(0,s1):
        for k2 in range(0,s2):
            x_min=max([-gauss_box_width,-k1]);
            x_max=min([gauss_box_width,(s1-k1-1)]);
            y_min=max([-gauss_box_width,-k2]);
            y_max=min([gauss_box_width,(s2-k2-1)]);
            [xx,yy]=np.meshgrid(range(y_min,y_max+1),range(x_min,x_max+1));
            gaus=fun_2d_gaussian(xx,yy,gauss_cell)   
            temp=np.corrcoef(gaus.reshape([1,gaus.size]),img[k1+x_min:k1+x_max+1,k2+y_min:k2+y_max+1].reshape([1,gaus.size]));
            new_deconv[k1,k2]=temp[0,1]
            
                
                        
    image_path="/Frame_%04i/Channel_Current" % (img_num)
    for x in range(0,filter_num+2): 
        image_path="%s/Filter_Step_%04i" % (image_path,x)
    
    try:
        main_h5_handle.__delitem__(image_path)
    except:
        temp=1;
        
    image_path="%s/Filtered_Image" % (image_path)
    main_h5_handle[image_path]=new_deconv;
    h5_image_new = main_h5_handle.get(image_path);
    h5_new_attrs = h5_image_new.attrs
    h5_new_attrs["Filter_Name"]="2D Correlation";
    h5_new_attrs["Number_Of_Variables"]=2;
    h5_new_attrs["Variable_1_Name"]="Filter width";
    h5_new_attrs["Variable_1_Value"]=gauss_box_width;
    h5_new_attrs["Variable_2_Name"]="Gassian width";
    h5_new_attrs["Variable_2_Value"]=gauss_width;
    
    
    h5_new_attrs["Spectroscopic_Indices"]=sec_i_ref;
    h5_new_attrs["Spectroscopic_Indices_Region"]=sec_i_reg;
    h5_new_attrs["Spectroscopic_Values"]=sec_v_ref;
    h5_new_attrs["Spectroscopic_Values_Region"]=sec_v_reg;
    h5_new_attrs["Position_Indices"]=pos_i_ref;
    h5_new_attrs["Position_Indices_Region"]=pos_i_reg;
    h5_new_attrs["Position_Values"]=pos_v_ref;
    h5_new_attrs["Position_Values_Region"]=pos_v_reg;
    h5_new_attrs["Parent"]=current_ref;
    h5_new_attrs["Parent_Region"]=current_reg;
    
    main_h5_handle.close()

    return 1
    
    
def fun_2d_gaussian(x,y,parm):
    
    import numpy as np   
    
    amp=np.double(parm[0]);
    
    x_cent=np.double(parm[1]);
    y_cent=np.double(parm[2]);
    
    x_wid=np.double(parm[3]);
    y_wid=np.double(parm[4]);
    
    ang=np.double(parm[5]);
    
    a = ((np.cos(ang)**2)/(2*(x_wid)**2))+((np.sin(ang)**2)/(2*(y_wid)**2));
    b = -((np.sin(2*ang))/(4*(x_wid)**2))+((np.sin(2*ang))/(4*(y_wid)**2));
    c =((np.sin(ang)**2)/(2*(x_wid)**2))+((np.cos(ang)**2)/(2*(y_wid)**2));
    
    gaussian=amp*(np.exp(-((a*(x-x_cent)**2)+(2*b*(x-x_cent)*(y-y_cent))+(c*(y-y_cent)**2))));

    return gaussian

def apply_invert_filter(file_in_h5,img_num,filter_num):
    
    import numpy as np
    import h5py as h5
   
   
    main_h5_handle = h5.File(file_in_h5,'r+')
    image_path="/Frame_%04i/Channel_Current" % (img_num)
    for x in range(0,filter_num+1): 
        image_path="%s/Filter_Step_%04i" % (image_path,x)        
    image_path="%s/Filtered_Image" % (image_path)
    
    h5_image=main_h5_handle.get(image_path)
    img = np.empty(h5_image.shape, dtype=h5_image.dtype)
    h5_image.read_direct(img)
    
    sec_i_ref=h5_image.attrs.get("Spectroscopic_Indices");
    sec_i_reg=h5_image.attrs.get("Spectroscopic_Indices_Region");
    sec_v_ref=h5_image.attrs.get("Spectroscopic_Values");
    sec_v_reg=h5_image.attrs.get("Spectroscopic_Values_Region");
    pos_i_ref=h5_image.attrs.get("Position_Indices");
    pos_i_reg=h5_image.attrs.get("Position_Indices_Region");
    pos_v_ref=h5_image.attrs.get("Position_Values");
    pos_v_reg=h5_image.attrs.get("Position_Values_Region");
    current_ref=h5_image.ref
    current_reg=h5_image.regionref[0:len(img)]
    
    m_img=img.mean();
    img=img-m_img;
    img=-img;
    img=img+m_img;
    

    image_path="/Frame_%04i/Channel_Current" % (img_num)
    for x in range(0,filter_num+2): 
        image_path="%s/Filter_Step_%04i" % (image_path,x)
    
    temp=0
    try:
        main_h5_handle.__delitem__(image_path)
    except:
        temp=1;
        
    image_path="%s/Filtered_Image" % (image_path)
    main_h5_handle[image_path]=img;
    h5_image_new = main_h5_handle.get(image_path);
    h5_new_attrs = h5_image_new.attrs
    h5_new_attrs["Filter_Name"]="invert z contrast";
    h5_new_attrs["Number_Of_Variables"]=0;
    
    
    h5_new_attrs["Spectroscopic_Indices"]=sec_i_ref;
    h5_new_attrs["Spectroscopic_Indices_Region"]=sec_i_reg;
    h5_new_attrs["Spectroscopic_Values"]=sec_v_ref;
    h5_new_attrs["Spectroscopic_Values_Region"]=sec_v_reg;
    h5_new_attrs["Position_Indices"]=pos_i_ref;
    h5_new_attrs["Position_Indices_Region"]=pos_i_reg;
    h5_new_attrs["Position_Values"]=pos_v_ref;
    h5_new_attrs["Position_Values_Region"]=pos_v_reg;
    h5_new_attrs["Parent"]=current_ref;
    h5_new_attrs["Parent_Region"]=current_reg;
    
    main_h5_handle.close
    
    return 1
    
def apply_find (file_path_h5,file_name_h5,file_path_png,file_name_png,filter_width,img_num,filter_num):
    
    import numpy as np
    import h5py as h5
    
    
    
    image_path="/Frame_%04i/Filtered_Data/Stack_0000" % (img_num)
    for x in range(0,filter_num+1): 
        image_path="%s/Filter_Step_%04i" % (image_path,x)
        

    s=1.1 # constant for now change latter 
    
    size=img.shape 
    s1=size[0]
    s2=size[1]
    
    mat_large=np.empty([s1+2*filter_width+1,s2+2*filter_width+1])
    mat_large[:,:]=np.inf
    
    for k1 in range(-filter_width,filter_width+1):
        for k2 in range(-filter_width,filter_width+1):
            mat_large[filter_width-k1:-(filter_width+k1)-1,filter_width-k2:-(filter_width+k2)-1]=np.minimum(mat_large[filter_width-k1:-filter_width-k1-1,filter_width-k2:-filter_width-k2-1],h5_image)
 
    deconv_mat_temp=mat_large[filter_width:len(mat_larg[1,:])-filter_width,filter_width:len(mat_larg[:,1])-filter_width];
    filtered_image=h5_image-deconv_mat_temp;

    return 1
    
def apply_binarization_filter(file_in_h5,img_num,filter_num):
    
    import numpy as np
    import h5py as h5
       
    main_h5_handle = h5.File(file_in_h5,'r+')
    image_path="/Frame_%04i/Channel_Current" % (img_num)
    for x in range(0,filter_num+1): 
        image_path="%s/Filter_Step_%04i" % (image_path,x)        
    image_path="%s/Filtered_Image" % (image_path)
    
    h5_image=main_h5_handle.get(image_path)
    img = np.empty(h5_image.shape, dtype=h5_image.dtype)
    h5_image.read_direct(img)
    
    sec_i_ref=h5_image.attrs.get("Spectroscopic_Indices");
    sec_i_reg=h5_image.attrs.get("Spectroscopic_Indices_Region");
    sec_v_ref=h5_image.attrs.get("Spectroscopic_Values");
    sec_v_reg=h5_image.attrs.get("Spectroscopic_Values_Region");
    pos_i_ref=h5_image.attrs.get("Position_Indices");
    pos_i_reg=h5_image.attrs.get("Position_Indices_Region");
    pos_v_ref=h5_image.attrs.get("Position_Values");
    pos_v_reg=h5_image.attrs.get("Position_Values_Region");
    current_ref=h5_image.ref
    current_reg=h5_image.regionref[0:len(img)]
    
    spec_ind=main_h5_handle[sec_i_ref];
    spec_ind=spec_ind[sec_i_reg];
    
    posi_ind=main_h5_handle[pos_i_ref];
    posi_ind=posi_ind[pos_i_reg];
    
    i_max=img.max()
    i_min=img.min()
    i_diff=i_max-i_min
    
    max_r=49
    filter_img=np.zeros([max_r+1,len(img)]) 
    time_out=np.zeros([max_r+1,1])
    time_out_i=np.zeros([max_r+1,1])

    for x in range(0,max_r+1):
        temp=np.zeros([len(img),1])
        r=x/max_r
        temp[img>(i_min+(i_diff*r))]=1
        filter_img[x,:]=temp[:,0];
        time_out[x]=r;
        time_out_i[x]=x+1;
        
            
    image_path="/Frame_%04i/Channel_Current" % (img_num)
    for x in range(0,filter_num+2): 
        image_path="%s/Filter_Step_%04i" % (image_path,x)
    
    try:
        main_h5_handle.__delitem__(image_path)
    except:
        temp=1;
    
    
    image_path_f="%s/Filtered_Image" % (image_path)
    main_h5_handle[image_path_f]=img;
    h5_image_new = main_h5_handle.get(image_path_f);
    h5_new_attrs = h5_image_new.attrs
    h5_new_attrs["Filter_Name"]="Threshold_Reference_Image";
    h5_new_attrs["Number_Of_Variables"]=0;
    
    
    h5_new_attrs["Spectroscopic_Indices"]=sec_i_ref;
    h5_new_attrs["Spectroscopic_Indices_Region"]=sec_i_reg;
    h5_new_attrs["Spectroscopic_Values"]=sec_v_ref;
    h5_new_attrs["Spectroscopic_Values_Region"]=sec_v_reg;
    h5_new_attrs["Position_Indices"]=pos_i_ref;
    h5_new_attrs["Position_Indices_Region"]=pos_i_reg;
    h5_new_attrs["Position_Values"]=pos_v_ref;
    h5_new_attrs["Position_Values_Region"]=pos_v_reg;
    h5_new_attrs["Parent"]=current_ref;
    h5_new_attrs["Parent_Region"]=current_reg;
    
    image_path_sv="%s/Spectroscopic_Values" % (image_path)
    main_h5_handle[image_path_sv]=time_out;
    h5_image_new = main_h5_handle.get(image_path_sv);
    new_sv_ref=h5_image_new.ref
    new_sv_reg=h5_image_new.regionref[0:len(time_out)]
    
    image_path_si="%s/Spectroscopic_Indices" % (image_path)
    main_h5_handle[image_path_si]=time_out_i;
    h5_image_new = main_h5_handle.get(image_path_si);
    new_si_ref=h5_image_new.ref
    new_si_reg=h5_image_new.regionref[0:len(time_out)]
    
    image_path_b="%s/Binary_Matrix" % (image_path)
    main_h5_handle[image_path_b]=filter_img;
    h5_image_new = main_h5_handle.get(image_path_b);
    h5_new_attrs = h5_image_new.attrs
    h5_new_attrs["Filter_Name"]="Threshold_Matrix";
    h5_new_attrs["Number_Of_Variables"]=0;
    
    
    h5_new_attrs["Spectroscopic_Indices"]=new_si_ref;
    h5_new_attrs["Spectroscopic_Indices_Region"]=new_si_reg;
    h5_new_attrs["Spectroscopic_Values"]=new_sv_ref;
    h5_new_attrs["Spectroscopic_Values_Region"]=new_sv_reg;
    h5_new_attrs["Position_Indices"]=pos_i_ref;
    h5_new_attrs["Position_Indices_Region"]=pos_i_reg;
    h5_new_attrs["Position_Values"]=pos_v_ref;
    h5_new_attrs["Position_Values_Region"]=pos_v_reg;
    h5_new_attrs["Parent"]=current_ref;
    h5_new_attrs["Parent_Region"]=current_reg;
    

    main_h5_handle.close()

    return 1
    
def apply_binarization_filter_select(file_in_h5,img_num,filter_num,threshold):
    
    import numpy as np
    import h5py as h5
       
    main_h5_handle = h5.File(file_in_h5,'r+')
    image_path="/Frame_%04i/Channel_Current" % (img_num)
    for x in range(0,filter_num+1): 
        image_path="%s/Filter_Step_%04i" % (image_path,x)        
    image_path="%s/Filtered_Image" % (image_path)
    
    h5_image=main_h5_handle.get(image_path)
    img = np.empty(h5_image.shape, dtype=h5_image.dtype)
    h5_image.read_direct(img)
    
    sec_i_ref=h5_image.attrs.get("Spectroscopic_Indices");
    sec_i_reg=h5_image.attrs.get("Spectroscopic_Indices_Region");
    sec_v_ref=h5_image.attrs.get("Spectroscopic_Values");
    sec_v_reg=h5_image.attrs.get("Spectroscopic_Values_Region");
    pos_i_ref=h5_image.attrs.get("Position_Indices");
    pos_i_reg=h5_image.attrs.get("Position_Indices_Region");
    pos_v_ref=h5_image.attrs.get("Position_Values");
    pos_v_reg=h5_image.attrs.get("Position_Values_Region");
    current_ref=h5_image.ref
    current_reg=h5_image.regionref[0:len(img)]
    
    spec_ind=main_h5_handle[sec_i_ref];
    spec_ind=spec_ind[sec_i_reg];
    
    posi_ind=main_h5_handle[pos_i_ref];
    posi_ind=posi_ind[pos_i_reg];
    
    i_max=img.max()
    i_min=img.min()
    i_diff=i_max-i_min
    
    max_r=49.0
    filter_img=np.zeros([1,len(img)])
    temp=np.zeros([len(img),1])
    r=threshold/max_r
    temp[img>(i_min+(i_diff*r))]=1
    filter_img=temp[:,0];
        
            
    image_path="/Frame_%04i/Channel_Current" % (img_num)
    for x in range(0,filter_num+2): 
        image_path="%s/Filter_Step_%04i" % (image_path,x)
    
    try:
        main_h5_handle.__delitem__(image_path)
    except:
        temp=1;
    
    
    image_path="%s/Filtered_Image" % (image_path)
    main_h5_handle[image_path]=filter_img;
    h5_image_new = main_h5_handle.get(image_path);
    h5_new_attrs = h5_image_new.attrs
    h5_new_attrs["Filter_Name"]="Threshold";
    h5_new_attrs["Number_Of_Variables"]=1;
    h5_new_attrs["Variable_1_Name"]="Threshold";
    h5_new_attrs["Variable_1_Value"]=r;
    
    
    h5_new_attrs["Spectroscopic_Indices"]=sec_i_ref;
    h5_new_attrs["Spectroscopic_Indices_Region"]=sec_i_reg;
    h5_new_attrs["Spectroscopic_Values"]=sec_v_ref;
    h5_new_attrs["Spectroscopic_Values_Region"]=sec_v_reg;
    h5_new_attrs["Position_Indices"]=pos_i_ref;
    h5_new_attrs["Position_Indices_Region"]=pos_i_reg;
    h5_new_attrs["Position_Values"]=pos_v_ref;
    h5_new_attrs["Position_Values_Region"]=pos_v_reg;
    h5_new_attrs["Parent"]=current_ref;
    h5_new_attrs["Parent_Region"]=current_reg;
    

    main_h5_handle.close()

    return


def cluster_into_atomic_columns(file_in_h5,img_num,filter_num,dist_val):
    
    import numpy as np      
    import h5py as h5
    
    main_h5_handle = h5.File(file_in_h5,'r+')
    image_path="/Frame_%04i/Channel_Current" % (img_num)
    for x in range(0,filter_num+1): 
        image_path="%s/Filter_Step_%04i" % (image_path,x)        
    image_path="%s/Filtered_Image" % (image_path)
    
    h5_image=main_h5_handle.get(image_path)
    img2 = np.empty(h5_image.shape, dtype=h5_image.dtype)
    h5_image.read_direct(img2)
    
    sec_i_ref=h5_image.attrs.get("Spectroscopic_Indices");
    sec_i_reg=h5_image.attrs.get("Spectroscopic_Indices_Region");
    sec_v_ref=h5_image.attrs.get("Spectroscopic_Values");
    sec_v_reg=h5_image.attrs.get("Spectroscopic_Values_Region");
    pos_i_ref=h5_image.attrs.get("Position_Indices");
    pos_i_reg=h5_image.attrs.get("Position_Indices_Region");
    pos_v_ref=h5_image.attrs.get("Position_Values");
    pos_v_reg=h5_image.attrs.get("Position_Values_Region");
    current_ref_m=h5_image.ref
    current_reg_m=h5_image.regionref[0:len(img2)]
    
    spec_ind=main_h5_handle[sec_i_ref];
    spec_ind=spec_ind[sec_i_reg];
    
    posi_ind=main_h5_handle[pos_i_ref];
    posi_ind=posi_ind[pos_i_reg];
    img = np.empty([max(posi_ind[:,0]),max(posi_ind[:,1])],dtype=h5_image.dtype)
    try:
        img[posi_ind[:,0]-1,posi_ind[:,1]-1]=img2[0:len(img2),0]
    except:
        img[posi_ind[:,0]-1,posi_ind[:,1]-1]=img2[0:len(img2)]

    centers=cluster_2d_oleg_return_geo_center(img,dist_val)

    image_path_org="/Frame_%04i/Channel_Current" % (img_num)
    image_path_new="/Frame_%04i/Channel_Finished" % (img_num)
    
    
    try:
        main_h5_handle.__delitem__(image_path_new)
    except:
        temp=1;
        
    old_folder=main_h5_handle.get(image_path_org)
    main_h5_handle.create_group(image_path_new)
    new_folder=main_h5_handle.get(image_path_new)
    new_folder_attrs = new_folder.attrs

    parrent_ref=old_folder.attrs.get("Origin");
    parrent_reg=old_folder.attrs.get("Origin_Region");
    parrent_name=old_folder.attrs.get("Origin_Name");
    new_folder_attrs["Origin"]=parrent_ref;
    new_folder_attrs["Origin_Region"]=parrent_reg;
    new_folder_attrs["Origin_Name"]=parrent_name;
    
    
    for x in range(0,filter_num+1): 
        image_path_org="%s/Filter_Step_%04i" % (image_path_org,x)   
        image_path_new="%s/Filter_Step_%04i" % (image_path_new,x)        
        image_path_org_temp="%s/Filtered_Image" % (image_path_org)       
        image_path_new_temp="%s/Filtered_Image" % (image_path_new)
        
        h5_image_old=main_h5_handle.get(image_path_org_temp)
        img_old = np.empty(h5_image_old.shape, dtype=h5_image_old.dtype)
        h5_image_old.read_direct(img_old)
        
        sec_i_ref=h5_image_old.attrs.get("Spectroscopic_Indices");
        sec_i_reg=h5_image_old.attrs.get("Spectroscopic_Indices_Region");
        sec_v_ref=h5_image_old.attrs.get("Spectroscopic_Values");
        sec_v_reg=h5_image_old.attrs.get("Spectroscopic_Values_Region");
        pos_i_ref=h5_image_old.attrs.get("Position_Indices");
        pos_i_reg=h5_image_old.attrs.get("Position_Indices_Region");
        pos_v_ref=h5_image_old.attrs.get("Position_Values");
        pos_v_rexg=h5_image_old.attrs.get("Position_Values_Region");
        filter_name=h5_image_old.attrs.get("Filter_Name");
        number_var=h5_image_old.attrs.get("Number_Of_Variables");
        
        main_h5_handle[image_path_new_temp]=img_old;
        h5_image_new = main_h5_handle.get(image_path);
        h5_new_attrs = h5_image_new.attrs
        h5_new_attrs["Filter_Name"]=filter_name;
        h5_new_attrs["Number_Of_Variables"]=number_var;
        for x in range(1,number_var+1):
            var_name=h5_image_old.attrs.get("Variable_%01i_Name" % (x));
            var_value=h5_image_old.attrs.get("Variable_%01i_Value" % (x));
            h5_new_attrs["Variable_%01i_Name" % (x)]=var_name;
            h5_new_attrs["Variable_%01i_Value" % (x)]=var_value;
    
        
        h5_new_attrs["Spectroscopic_Indices"]=sec_i_ref;
        h5_new_attrs["Spectroscopic_Indices_Region"]=sec_i_reg;
        h5_new_attrs["Spectroscopic_Values"]=sec_v_ref;
        h5_new_attrs["Spectroscopic_Values_Region"]=sec_v_reg;
        h5_new_attrs["Position_Indices"]=pos_i_ref;
        h5_new_attrs["Position_Indices_Region"]=pos_i_reg;
        h5_new_attrs["Position_Values"]=pos_v_ref;
        h5_new_attrs["Position_Values_Region"]=pos_v_reg;
        h5_new_attrs["Parent"]=parrent_ref;
        h5_new_attrs["Parent_Region"]=parrent_reg;
        
        parrent_ref=h5_image_new.ref
        parrent_reg=h5_image_new.regionref[0:len(img_old)];
        
    image_path="%s/Lattice/Positions" % (image_path_new)
    main_h5_handle[image_path]=centers;
    h5_image_new = main_h5_handle.get(image_path);
    h5_new_attrs = h5_image_new.attrs
    h5_new_attrs["Method"]="Density_Clustering";
    h5_new_attrs["Number_Of_Variables"]=1;
    h5_new_attrs["Variable_1_Name"]="Connection_Distance";
    h5_new_attrs["Variable_1_Value"]=dist_val;
    h5_new_attrs["Parent"]=current_ref_m;
    h5_new_attrs["Parent_Region"]=current_reg_m;
    
    main_h5_handle.close()
        
    
def cluster_2d_oleg(mat_in,dist_val):

    import numpy as np
    
    tt=[0,0]
    clusters=[];
    to_cluster=np.argwhere(mat_in);
    to_cluster=to_cluster.tolist()

    while (len(to_cluster)>0):
        clust=[];
        final_clust=[];
        clust.append(to_cluster[0]);
        to_cluster.remove(to_cluster[0])
        while ((len(clust)>0) & (len(to_cluster)>0)):
            tt[0]=5000.0 * dist_val;
            tt[1]=len(to_cluster);
            t1=min(tt);
            t1=int(t1);
            to_cluster_t=to_cluster[0:t1];
            dem_diff=abs(to_cluster_t- np.tile(clust[0],[t1,1]));
            diff_vec=np.argwhere(sum(dem_diff,1)<=dist_val);
            for px in range(len(diff_vec),0,-1):
                pt=diff_vec[px-1];
                clust.append(to_cluster[pt]);
                to_cluster.remove(to_cluster[pt])
            final_clust.append(clust[0]);
            clust.remove(clust[0]);
    
        if (len(clust)>0):
            while (len(clust)<0):
                final_clust.append(clust[0]);
                clust.remove(clust[0]);
            
        clusters.append(final_clust);
        
        
    return clusters

def cluster_2d_oleg_return_geo_center(mat_in,dist_val):
    
    import numpy as np
    
    tt=[0,0]
    clusters=[];
    to_cluster=np.argwhere(mat_in);
    to_cluster=to_cluster.tolist()

    while (len(to_cluster)>0):
        clust=[];
        final_clust=[];
        clust.append(to_cluster[0]);
        to_cluster.remove(to_cluster[0])
        while ((len(clust)>0) & (len(to_cluster)>0)):
            tt[0]=5000.0 * dist_val;
            tt[1]=len(to_cluster);
            t1=min(tt);
            t1=int(t1);
            to_cluster_t=to_cluster[0:t1];
            dem_diff=abs(to_cluster_t- np.tile(clust[0],[t1,1]));         
            diff_vec=np.argwhere(np.sum(dem_diff,axis=1)<=dist_val); 
            for px in range(len(diff_vec),0,-1):
                pt=diff_vec[px-1];
                clust.append(to_cluster[pt]);
                to_cluster.remove(to_cluster[pt])
            final_clust.append(clust[0]);
            clust.remove(clust[0]);
    
        if (len(clust)>0):
            while (len(clust)<0):
                final_clust.append(clust[0]);
                clust.remove(clust[0]);

        if (len(final_clust)>2):
            clusters.append(np.mean(final_clust,0));
        
        
    clusters=np.array(clusters)  
    clusters=clusters.tolist();
    
    return clusters


def return_img(file_in_h5,img_num,filter_num):
    
    import numpy as np
    import h5py as h5
    
       
    main_h5_handle = h5.File(file_in_h5,'r+')
    image_path="/Frame_%04i/Channel_Current" % (img_num)
    for x in range(0,filter_num+1): 
        image_path="%s/Filter_Step_%04i" % (image_path,x)        
    image_path="%s/Filtered_Image" % (image_path)
    
    h5_image=main_h5_handle.get(image_path)
    img2 = np.empty(h5_image.shape, dtype=h5_image.dtype)
    h5_image.read_direct(img2)
    
    sec_i_ref=h5_image.attrs.get("Spectroscopic_Indices");
    sec_i_reg=h5_image.attrs.get("Spectroscopic_Indices_Region");
    sec_v_ref=h5_image.attrs.get("Spectroscopic_Values");
    sec_v_reg=h5_image.attrs.get("Spectroscopic_Values_Region");
    pos_i_ref=h5_image.attrs.get("Position_Indices");
    pos_i_reg=h5_image.attrs.get("Position_Indices_Region");
    pos_v_ref=h5_image.attrs.get("Position_Values");
    pos_v_reg=h5_image.attrs.get("Position_Values_Region");
    
    spec_ind = main_h5_handle[sec_i_ref];
    spec_ind = spec_ind[sec_i_reg];
    
    posi_ind = main_h5_handle[pos_i_ref];
    posi_ind = posi_ind[pos_i_reg];
    img = np.empty([max(posi_ind[:, 0]), max(posi_ind[:, 1])], dtype=h5_image.dtype)
    try:
        img[posi_ind[:, 0]-1, posi_ind[:, 1]-1] = img2[0:len(img2), 0]
    except:
        img[posi_ind[:, 0]-1, posi_ind[:, 1]-1] = img2[0:len(img2)]
    main_h5_handle.close()
        

    return img
    
def return_pos(file_in_h5,img_num):
    
    import numpy as np
    import h5py as h5
    
       
    main_h5_handle = h5.File(file_in_h5,'r+')
    image_path="/Frame_%04i/Channel_Finished" % (img_num)
    type_ref=type(main_h5_handle.get(image_path));
    temp=1
    x=-1;
    while temp: 
        x=x+1
        image_path="%s/Filter_Step_%04i" % (image_path,x)
        image_temp="%s/Lattice" % (image_path)
        temp_loc=main_h5_handle.get(image_temp)
        if type(temp_loc)==type_ref:
            temp=0;

                    
    image_path=image_temp     
    image_path="%s/Positions" % (image_path)
    
    h5_image=main_h5_handle.get(image_path)
    pos = np.empty(h5_image.shape, dtype=h5_image.dtype)
    h5_image.read_direct(pos)
    
    return pos
    
def run_PCA_atoms(file_in_h5,img_num,box_width):
    
    import numpy as np
    import h5py as h5
    from scipy import linalg
    
       
    main_h5_handle = h5.File(file_in_h5,'r+')
    image_path="/Frame_%04i/Channel_Finished" % (img_num)
    type_ref=type(main_h5_handle.get(image_path));
    temp=1
    x=-1;
    while temp: 
        x=x+1
        image_path="%s/Filter_Step_%04i" % (image_path,x)
        image_temp="%s/Lattice" % (image_path)
        temp_loc=main_h5_handle.get(image_temp)
        if type(temp_loc)==type_ref:
            temp=0;

                    
    pos_path=image_temp     
    image_path="%s/Positions" % (pos_path)
    
    h5_image=main_h5_handle.get(image_path)
    pos = np.empty(h5_image.shape, dtype=h5_image.dtype)
    h5_image.read_direct(pos)
    current_ref=h5_image.ref
    current_reg=h5_image.regionref[0:len(pos),0:len(pos[0])]
    
    image_path="/Frame_%04i/Channel_Current" % (img_num)
    image_path="%s/Filter_Step_%04i" % (image_path,0)        
    image_path="%s/Filtered_Image" % (image_path)
    
    h5_image=main_h5_handle.get(image_path)
    img2 = np.empty(h5_image.shape, dtype=h5_image.dtype)
    h5_image.read_direct(img2)
    
    sec_i_ref=h5_image.attrs.get("Spectroscopic_Indices");
    sec_i_reg=h5_image.attrs.get("Spectroscopic_Indices_Region");
    pos_i_ref=h5_image.attrs.get("Position_Indices");
    pos_i_reg=h5_image.attrs.get("Position_Indices_Region");
    
    spec_ind=main_h5_handle[sec_i_ref];
    spec_ind=spec_ind[sec_i_reg];
    
    posi_ind=main_h5_handle[pos_i_ref];
    posi_ind=posi_ind[pos_i_reg];
    img = np.empty([max(posi_ind[:,0]),max(posi_ind[:,1])],dtype=h5_image.dtype)
    try:
        img[posi_ind[:,0]-1,posi_ind[:,1]-1]=img2[0:len(img2),0]
    except:
        img[posi_ind[:,0]-1,posi_ind[:,1]-1]=img2[0:len(img2)]
        
        
    sel_vec=np.zeros([len(pos[:,0]),1])
    kk=0;
    new_pos=[];
    img_vectors=[];
    for k1 in range(0,len(pos[:,0])):
        if (pos[k1,0].round()>=box_width and pos[k1,1].round()>=box_width and pos[k1,0].round()<=len(img[:,0])-box_width and pos[k1,1].round()<=len(img[0,:])-box_width):
            sel_vec[k1]=1
            new_pos.append(pos[k1,:])
            vector=img[pos[k1,0]-box_width:pos[k1,0]+box_width,pos[k1,1]-box_width:pos[k1,1]+box_width];
            img_vectors.append(vector.reshape([(box_width*2)**2]))

    new_pos=array(new_pos)
    img_vectors=array(img_vectors)
    
    U, S, V = linalg.svd(img_vectors)
    V=V[0:len(S)]
    
    temp=1
    x=-1;

    while temp: 
        x=x+1
        image_temp="%s/Analysis_%04i" % (pos_path,x)
        temp_loc=main_h5_handle.get(image_temp)
        if type(temp_loc)!=type_ref:
            temp=0;
    
    image_path=image_temp;
    
    

    
    [xx,yy]=np.meshgrid(range(1,len(U)+1),range(1,len(U[0])+1))
    xx=np.reshape(xx,[1,len(U)*len(U[0])])
    yy=np.reshape(yy,[1,len(U)*len(U[0])])
    xy=[xx,yy]
    U=np.reshape(U,[1,len(U)*len(U[0])])
    
    image_path_sv="%s/Spectroscopic_Values_01" % (image_path)
    main_h5_handle[image_path_sv]=[1];
    h5_image_new = main_h5_handle.get(image_path_sv);
    new_sv_ref=h5_image_new.ref
    new_sv_reg=h5_image_new.regionref[0]
    
    image_path_si="%s/Spectroscopic_Indices_01" % (image_path)
    main_h5_handle[image_path_si]=[1];
    h5_image_new = main_h5_handle.get(image_path_si);
    new_si_ref=h5_image_new.ref
    new_si_reg=h5_image_new.regionref[0]
    
    image_path_sv="%s/Position_Values_01" % (image_path)
    main_h5_handle[image_path_sv]=xy;
    h5_image_new = main_h5_handle.get(image_path_sv);
    new_pv_ref=h5_image_new.ref
    new_pv_reg=h5_image_new.regionref[0:1,0:len(xy[0])]
    
    image_path_si="%s/Position_Indices_01" % (image_path)
    main_h5_handle[image_path_si]=xy;
    h5_image_new = main_h5_handle.get(image_path_si);
    new_pi_ref=h5_image_new.ref
    new_pi_reg=h5_image_new.regionref[0:1,0:len(xy[0])]
    
    image_path_b="%s/Analysis_Data_01_U" % (image_path)
    main_h5_handle[image_path_b]=U;
    h5_image_new = main_h5_handle.get(image_path_b);
    h5_new_attrs = h5_image_new.attrs
    h5_new_attrs["Filter_Name"]="PCA_Atom_Shape";
    h5_new_attrs["Number_Of_Variables"]=1;
    h5_new_attrs["Variable_1_Name"]="Box Width";
    h5_new_attrs["Variable_1_Value"]=box_width;
    
    h5_new_attrs["Spectroscopic_Indices"]=new_si_ref;
    h5_new_attrs["Spectroscopic_Indices_Region"]=new_si_reg;
    h5_new_attrs["Spectroscopic_Values"]=new_sv_ref;
    h5_new_attrs["Spectroscopic_Values_Region"]=new_sv_reg;
    h5_new_attrs["Position_Indices"]=new_pi_ref;
    h5_new_attrs["Position_Indices_Region"]=new_pi_reg;
    h5_new_attrs["Position_Values"]=new_pv_ref;
    h5_new_attrs["Position_Values_Region"]=new_pv_reg;
    h5_new_attrs["Parent"]=current_ref;
    h5_new_attrs["Parent_Region"]=current_reg;
       
    
    [xy]=np.meshgrid(range(1,len(S)+1))

    image_path_sv="%s/Spectroscopic_Values_02" % (image_path)
    main_h5_handle[image_path_sv]=[1];
    h5_image_new = main_h5_handle.get(image_path_sv);
    new_sv_ref=h5_image_new.ref
    new_sv_reg=h5_image_new.regionref[0]
    
    image_path_si="%s/Spectroscopic_Indices_02" % (image_path)
    main_h5_handle[image_path_si]=[1];
    h5_image_new = main_h5_handle.get(image_path_si);
    new_si_ref=h5_image_new.ref
    new_si_reg=h5_image_new.regionref[0]
    
    image_path_sv="%s/Position_Values_02" % (image_path)
    main_h5_handle[image_path_sv]=xy;
    h5_image_new = main_h5_handle.get(image_path_sv);
    new_pv_ref=h5_image_new.ref
    new_pv_reg=h5_image_new.regionref[0:len(xy)]
    
    image_path_si="%s/Position_Indices_02" % (image_path)
    main_h5_handle[image_path_si]=xy;
    h5_image_new = main_h5_handle.get(image_path_si);
    new_pi_ref=h5_image_new.ref
    new_pi_reg=h5_image_new.regionref[0:len(xy)]
       
    image_path_b="%s/Analysis_Data_02_S" % (image_path)
    main_h5_handle[image_path_b]=S;
    h5_image_new = main_h5_handle.get(image_path_b);
    h5_new_attrs = h5_image_new.attrs
    h5_new_attrs["Filter_Name"]="PCA_Atom_Shape";
    h5_new_attrs["Number_Of_Variables"]=1;
    h5_new_attrs["Variable_1_Name"]="Box Width";
    h5_new_attrs["Variable_1_Value"]=box_width;
    
    h5_new_attrs["Spectroscopic_Indices"]=new_si_ref;
    h5_new_attrs["Spectroscopic_Indices_Region"]=new_si_reg;
    h5_new_attrs["Spectroscopic_Values"]=new_sv_ref;
    h5_new_attrs["Spectroscopic_Values_Region"]=new_sv_reg;
    h5_new_attrs["Position_Indices"]=new_pi_ref;
    h5_new_attrs["Position_Indices_Region"]=new_pi_reg;
    h5_new_attrs["Position_Values"]=new_pv_ref;
    h5_new_attrs["Position_Values_Region"]=new_pv_reg;
    h5_new_attrs["Parent"]=current_ref;
    h5_new_attrs["Parent_Region"]=current_reg;
       
       
     
    [xx,yy]=np.meshgrid(range(1,len(V)+1),range(1,len(V[0])+1))
    xx=np.reshape(xx,[1,len(V)*len(V[0])])
    yy=np.reshape(yy,[1,len(V)*len(V[0])])
    xy=[xx,yy]
    V=np.reshape(U,[1,len(U)*len(U[0])])
    

    image_path_sv="%s/Spectroscopic_Values_03" % (image_path)
    main_h5_handle[image_path_sv]=[1];
    h5_image_new = main_h5_handle.get(image_path_sv);
    new_sv_ref=h5_image_new.ref
    new_sv_reg=h5_image_new.regionref[0]
    
    image_path_si="%s/Spectroscopic_Indices_03" % (image_path)
    main_h5_handle[image_path_si]=[1];
    h5_image_new = main_h5_handle.get(image_path_si);
    new_si_ref=h5_image_new.ref
    new_si_reg=h5_image_new.regionref[0]
    
    image_path_sv="%s/Position_Values_03" % (image_path)
    main_h5_handle[image_path_sv]=xy;
    h5_image_new = main_h5_handle.get(image_path_sv);
    new_pv_ref=h5_image_new.ref
    new_pv_reg=h5_image_new.regionref[0:1,0:len(xy[0])]
    
    image_path_si="%s/Position_Indices_03" % (image_path)
    main_h5_handle[image_path_si]=xy;
    h5_image_new = main_h5_handle.get(image_path_si);
    new_pi_ref=h5_image_new.ref
    new_pi_reg=h5_image_new.regionref[0:1,0:len(xy[0])]
    
    image_path_b="%s/Analysis_Data_03_V" % (image_path)
    main_h5_handle[image_path_b]=V1;
    h5_image_new = main_h5_handle.get(image_path_b);
    h5_new_attrs = h5_image_new.attrs
    h5_new_attrs["Filter_Name"]="PCA_Atom_Shape";
    h5_new_attrs["Number_Of_Variables"]=1;
    h5_new_attrs["Variable_1_Name"]="Box Width";
    h5_new_attrs["Variable_1_Value"]=box_width;
    
    
    h5_new_attrs["Spectroscopic_Indices"]=new_si_ref;
    h5_new_attrs["Spectroscopic_Indices_Region"]=new_si_reg;
    h5_new_attrs["Spectroscopic_Values"]=new_sv_ref;
    h5_new_attrs["Spectroscopic_Values_Region"]=new_sv_reg;
    h5_new_attrs["Position_Indices"]=new_pi_ref;
    h5_new_attrs["Position_Indices_Region"]=new_pi_reg;
    h5_new_attrs["Position_Values"]=new_pv_ref;
    h5_new_attrs["Position_Values_Region"]=new_pv_reg;
    h5_new_attrs["Parent"]=current_ref;
    h5_new_attrs["Parent_Region"]=current_reg;
    

    main_h5_handle.close()
    
    return 




# -*- coding: utf-8 -*-
"""
Created on Thu May 05 13:29:12 2016

@author: Suhas Somnath
"""
# TODO: All general plotting functions should support data with 1, 2, or 3 spatial dimensions.

from __future__ import division, print_function, absolute_import, unicode_literals

import inspect
from warnings import warn
import sys
import h5py
import matplotlib.pyplot as plt
import numpy as np
import scipy
from scipy.signal import blackman
from matplotlib.colors import LinearSegmentedColormap
from mpl_toolkits.axes_grid1 import ImageGrid

from ..io.hdf_utils import reshape_to_Ndims, get_formatted_labels


def set_tick_font_size(axes, font_size):
    """
    Sets the font size of the ticks in the provided axes

    Parameters
    ----------
    axes : matplotlib.pyplot.axis object or list of axis objects
        axes to set font sizes
    font_size : unigned int
        Font size
    """

    def __set_axis_tick(axis):
        """
        Sets the font sizes to the x and y axis in the given axis object

        Parameters
        ----------
        axis : matplotlib.pyplot.axis object
            axis to set font sizes
        """
        for tick in axis.xaxis.get_major_ticks():
            tick.label.set_fontsize(font_size)
        for tick in axis.yaxis.get_major_ticks():
            tick.label.set_fontsize(font_size)

    if hasattr(axes, '__iter__'):
        for axis in axes:
            __set_axis_tick(axis)
    else:
        __set_axis_tick(axes)


def cmap_jet_white_center():
    """
    Generates the jet colormap with a white center

    Returns
    -------
    white_jet : matplotlib.colors.LinearSegmentedColormap object
        color map object that can be used in place of plt.cm.viridis
    """
    # For red - central column is like brightness
    # For blue - last column is like brightness
    cdict = {'red': ((0.00, 0.0, 0.0),
                     (0.30, 0.0, 0.0),
                     (0.50, 1.0, 1.0),
                     (0.90, 1.0, 1.0),
                     (1.00, 0.5, 1.0)),
             'green': ((0.00, 0.0, 0.0),
                       (0.10, 0.0, 0.0),
                       (0.42, 1.0, 1.0),
                       (0.58, 1.0, 1.0),
                       (0.90, 0.0, 0.0),
                       (1.00, 0.0, 0.0)),
             'blue': ((0.00, 0.0, 0.5),
                      (0.10, 1.0, 1.0),
                      (0.50, 1.0, 1.0),
                      (0.70, 0.0, 0.0),
                      (1.00, 0.0, 0.0))
             }
    return LinearSegmentedColormap('white_jet', cdict)


def cmap_from_rgba(name, interp_vals, normalization_val):
    """
    Generates a colormap given a matlab-style interpolation table

    Parameters
    ----------
    name : String / Unicode
        Name of the desired colormap
    interp_vals : List of tuples
        Interpolation table that describes the desired color map. Each entry in the table should be described as:
        (position in the colorbar, (red, green, blue, alpha))
        The position in the color bar, red, green, blue, and alpha vary from 0 to the normalization value
    normalization_val : number
        The common maximum value for the position in the color bar, red, green, blue, and alpha

    Returns
    -------
    new_cmap : matplotlib.colors.LinearSegmentedColormap object
        desired color map
    """

    normalization_val = np.round(1.0 * normalization_val)

    cdict = {'red': tuple([(dist / normalization_val, colors[0] / normalization_val, colors[0] / normalization_val)
                           for (dist, colors) in interp_vals][::-1]),
             'green': tuple([(dist / normalization_val, colors[1] / normalization_val, colors[1] / normalization_val)
                             for (dist, colors) in interp_vals][::-1]),
             'blue': tuple([(dist / normalization_val, colors[2] / normalization_val, colors[2] / normalization_val)
                            for (dist, colors) in interp_vals][::-1]),
             'alpha': tuple([(dist / normalization_val, colors[3] / normalization_val, colors[3] / normalization_val)
                            for (dist, colors) in interp_vals][::-1])}

    return LinearSegmentedColormap(name, cdict)


def make_linear_alpha_cmap(name, solid_color, normalization_val, min_alpha=0, max_alpha=1):
    """
    Generates a transparent to opaque color map based on a single solid color

    Parameters
    ----------
    name : String / Unicode
        Name of the desired colormap
    solid_color : List of numbers
        red, green, blue, and alpha values for a specific color
    normalization_val : number
        The common maximum value for the red, green, blue, and alpha values. This is 1 in matplotlib
    min_alpha : float (optional. Default = 0 : ie- transparent)
        Lowest alpha value for the bottom of the color bar
    max_alpha : float (optional. Default = 1 : ie- opaque)
        Highest alpha value for the top of the color bar

    Returns
    -------
    new_cmap : matplotlib.colors.LinearSegmentedColormap object
        transparent to opaque color map based on the provided color
    """
    solid_color = np.array(solid_color) / normalization_val * 1.0
    interp_table = [(1.0, (solid_color[0], solid_color[1], solid_color[2], max_alpha)),
                    (0, (solid_color[0], solid_color[1], solid_color[2], min_alpha))]
    return cmap_from_rgba(name, interp_table, 1)


def cmap_hot_desaturated():
    """
    Returns a desaturated color map based on the hot colormap

    Returns
    -------
    new_cmap : matplotlib.colors.LinearSegmentedColormap object
        Desaturated version of the hot color map
    """
    hot_desaturated = [(255.0, (255, 76, 76, 255)),
                       (218.5, (107, 0, 0, 255)),
                       (182.1, (255, 96, 0, 255)),
                       (145.6, (255, 255, 0, 255)),
                       (109.4, (0, 127, 0, 255)),
                       (72.675, (0, 255, 255, 255)),
                       (36.5, (0, 0, 91, 255)),
                       (0, (71, 71, 219, 255))]

    return cmap_from_rgba('hot_desaturated', hot_desaturated, 255)


def discrete_cmap(num_bins, base_cmap=plt.cm.viridis):
    """
    Create an N-bin discrete colormap from the specified input map

    Parameters
    ----------
    num_bins : unsigned int
        Number of discrete bins
    base_cmap : matplotlib.colors.LinearSegmentedColormap object
        Base color map to discretize

    Returns
    -------
    new_cmap : matplotlib.colors.LinearSegmentedColormap object
        Discretized color map

    Notes
    -----
    Jake VanderPlas License: BSD-style
    https://gist.github.com/jakevdp/91077b0cae40f8f8244a

    """

    base = plt.cm.get_cmap(base_cmap)
    color_list = base(np.linspace(0, 1, num_bins))
    cmap_name = base.name + str(num_bins)
    return base.from_list(cmap_name, color_list, num_bins)


def _add_loop_parameters(axes, switching_coef_vec):
    """
    Add the loop parameters for the given loop to a list of axes

    Parameters
    ----------
    axes : list of matplotlib.pyplo.axes
        Plot axes to add the coeffients to
    switching_coef_vec : 1D numpy.ndarray
        Array of loop parameters arranged by position

    Returns
    -------
    axes : list of matplotlib.pyplo.axes
    """
    positions = np.linspace(0, switching_coef_vec.shape[0] - 1, len(axes.flat), dtype=np.int)

    for ax, pos in zip(axes.flat, positions):
        ax.axvline(switching_coef_vec[pos]['V+'], c='k', label='V+')
        ax.axvline(switching_coef_vec[pos]['V-'], c='r', label='V-')
        ax.axvline(switching_coef_vec[pos]['Nucleation Bias 1'], c='k', ls=':', label='Nucleation Bias 1')
        ax.axvline(switching_coef_vec[pos]['Nucleation Bias 2'], c='r', ls=':', label='Nucleation Bias 2')
        ax.axhline(switching_coef_vec[pos]['R+'], c='k', ls='-.', label='R+')
        ax.axhline(switching_coef_vec[pos]['R-'], c='r', ls='-.', label='R-')

    return axes

def rainbow_plot(ax, ao_vec, ai_vec, num_steps=32, cmap=plt.cm.viridis, **kwargs):
    """
    Plots the input against the output waveform (typically loops).
    The color of the curve changes as a function of time using the jet colorscheme

    Parameters
    ----------
    ax : axis handle
        Axis to plot the curve
    ao_vec : 1D float numpy array
        vector that forms the X axis
    ai_vec : 1D float numpy array
        vector that forms the Y axis
    num_steps : unsigned int (Optional)
        Number of discrete color steps
    cmap : matplotlib.colors.LinearSegmentedColormap object
        Colormap to be used
    """
    pts_per_step = int(len(ai_vec) / num_steps)
    for step in range(num_steps - 1):
        ax.plot(ao_vec[step * pts_per_step:(step + 1) * pts_per_step],
                ai_vec[step * pts_per_step:(step + 1) * pts_per_step],
                color=cmap(255 * step / num_steps), **kwargs)
    # plot the remainder:
    ax.plot(ao_vec[(num_steps - 1) * pts_per_step:],
            ai_vec[(num_steps - 1) * pts_per_step:],
            color=cmap(255 * num_steps / num_steps), **kwargs)
    """
    CS3=plt.contourf([[0,0],[0,0]], range(0,310),cmap=plt.cm.viridis)
    fig.colorbar(CS3)"""


def plot_line_family(axis, x_axis, line_family, line_names=None, label_prefix='Line', label_suffix='', cmap=plt.cm.viridis,
                     **kwargs):
    """
    Plots a family of lines with a sequence of colors

    Parameters
    ----------
    axis : axis handle
        Axis to plot the curve
    x_axis : array-like
        Values to plot against
    line_family : 2D numpy array
        family of curves arranged as [curve_index, features]
    line_names : array-like
        array of string or numbers that represent the identity of each curve in the family
    label_prefix : string / unicode
        prefix for the legend (before the index of the curve)
    label_suffix : string / unicode
        suffix for the legend (after the index of the curve)
    cmap : matplotlib.colors.LinearSegmentedColormap object
        Colormap to be used
    """
    num_lines = line_family.shape[0]

    if line_names is None:
        line_names = ['{} {} {}'.format(label_prefix, line_ind, label_suffix) for line_ind in range(num_lines)]
    else:
        if len(line_names) != num_lines:
            warn('Line names of different length compared to provided dataset')
            line_names = ['{} {} {}'.format(label_prefix, line_ind, label_suffix) for line_ind in range(num_lines)]

    for line_ind in range(num_lines):
        axis.plot(x_axis, line_family[line_ind],
                  label=line_names[line_ind],
                  color=cmap(int(255 * line_ind / (num_lines - 1))), **kwargs)


def plot_map(axis, data, stdevs=2, origin='lower', **kwargs):
    """
    Plots a 2d map with a tight z axis, with or without color bars.
    Note that the direction of the y axis is flipped if the color bar is required

    Parameters
    ----------
    axis : matplotlib.pyplot.axis object
        Axis to plot this map onto
    data : 2D real numpy array
        Data to be plotted
    stdevs : unsigned int (Optional. Default = 2)
        Number of standard deviations to consider for plotting
    origin : str
        Where should the origin of the image data be located.  'lower' sets the origin to the
        bottom left, 'upper' sets it to the upper left.
        Default 'lower'

    Returns
    -------
    """
    data_mean = np.mean(data)
    data_std = np.std(data)
    im = axis.imshow(data, interpolation='none',
                     vmin=data_mean - stdevs * data_std,
                     vmax=data_mean + stdevs * data_std,
                     origin=origin,
                     **kwargs)
    axis.set_aspect('auto')

    return im


def plot_loops(excit_wfm, datasets, line_colors=[], dataset_names=[], evenly_spaced=True, plots_on_side=5, x_label='',
               y_label='', subtitles='Position', title='', central_resp_size=None, use_rainbow_plots=False, h5_pos=None):
    # TODO: Allow multiple excitation waveforms
    """
    Plots loops from multiple datasets from up to 25 evenly spaced positions

    Parameters
    -----------
    excit_wfm : 1D numpy float array
        Excitation waveform in the time domain
    datasets : list of 2D numpy arrays or 2D hyp5.Dataset objects
        Datasets containing data arranged as (pixel, time)
    line_colors : list of strings
        Colors to be used for each of the datasets
    dataset_names : (Optional) list of strings
        Names of the different datasets to be compared
    h5_pos : HDF5 dataset reference or 2D numpy array
        Dataset containing position indices
    central_resp_size : (optional) unsigned integer
        Number of responce sample points from the center of the waveform to show in plots. Useful for SPORC
    evenly_spaced : boolean
        Evenly spaced positions or first N positions
    plots_on_side : unsigned int
        Number of plots on each side
    use_rainbow_plots : (optional) Boolean
        Plot the lines as a function of spectral index (eg. time)
    x_label : (optional) String
        X Label for all plots
    y_label : (optional) String
        Y label for all plots
    subtitles : (optional) String
        prefix for title over each plot
    title : (optional) String
        Main plot title

    Returns
    ---------
    fig, axes
    """
    if type(datasets) in [h5py.Dataset, np.ndarray]:
        # can be numpy array or h5py.dataset
        num_pos = datasets.shape[0]
        num_points = datasets.shape[1]
        datasets = [datasets]
        line_colors = ['b']
        dataset_names = ['Default']
    else:
        # First check if the datasets are correctly shaped:
        num_pos_es = list()
        num_points_es = list()
        for dataset in datasets:
            num_pos_es.append(dataset.shape[0])
            num_points_es.append(dataset.shape[1])
        num_pos_es = np.array(num_pos_es)
        num_points_es = np.array(num_points_es)
        if np.unique(num_pos_es).size > 1 or np.unique(num_points_es).size > 1:
            warn('Datasets of incompatible sizes')
            return
        num_pos = np.unique(num_pos_es)[0]
        num_points = np.unique(num_points_es)[0]

        # Next the identification of datasets:
        if len(dataset_names) > len(datasets):
            # remove additional titles
            dataset_names = dataset_names[:len(datasets)]
        elif len(dataset_names) < len(datasets):
            # add titles
            dataset_names = dataset_names + ['Dataset' + ' ' + str(x) for x in range(len(dataset_names), len(datasets))]
        if len(line_colors) != len(datasets):
            color_list = ['b', 'g', 'r', 'c', 'm', 'y', 'k', 'pink', 'brown', 'orange']
            if len(datasets) < len(color_list):
                remaining_colors = [x for x in color_list if x not in line_colors]
                line_colors += remaining_colors[:len(datasets) - len(color_list)]
            else:
                warn('Insufficient number of line colors provided')
                return


    if excit_wfm.size != num_points:
        warn('Length of excitation waveform not compatible with second axis of datasets')
        return

    plots_on_side = min(abs(plots_on_side), 5)

    sq_num_plots = min(plots_on_side, int(round(num_pos ** 0.5)))
    if evenly_spaced:
        chosen_pos = np.linspace(0, num_pos - 1, sq_num_plots ** 2, dtype=int)
    else:
        chosen_pos = np.arange(sq_num_plots ** 2, dtype=int)

    fig, axes = plt.subplots(nrows=sq_num_plots, ncols=sq_num_plots, sharex=True, figsize=(12, 12))
    axes_lin = axes.flatten()

    cent_ind = int(0.5 * excit_wfm.size)
    if central_resp_size:
        sz = int(0.5 * central_resp_size)
        l_resp_ind = cent_ind - sz
        r_resp_ind = cent_ind + sz
    else:
        l_resp_ind = 0
        r_resp_ind = excit_wfm.size

    for count, posn in enumerate(chosen_pos):
        if use_rainbow_plots and len(datasets) == 1:
            rainbow_plot(axes_lin[count], excit_wfm[l_resp_ind:r_resp_ind], datasets[0][posn, l_resp_ind:r_resp_ind])
        else:
            for dataset, col_val in zip(datasets, line_colors):
                axes_lin[count].plot(excit_wfm[l_resp_ind:r_resp_ind], dataset[posn, l_resp_ind:r_resp_ind], color=col_val)
        if h5_pos is not None:
            # print('Row ' + str(h5_pos[posn,1]) + ' Col ' + str(h5_pos[posn,0]))
            axes_lin[count].set_title('Row ' + str(h5_pos[posn, 1]) + ' Col ' + str(h5_pos[posn, 0]), fontsize=12)
        else:
            axes_lin[count].set_title(subtitles + ' ' + str(posn), fontsize=12)

        if count % sq_num_plots == 0:
            axes_lin[count].set_ylabel(y_label, fontsize=12)
        if count >= (sq_num_plots - 1) * sq_num_plots:
            axes_lin[count].set_xlabel(x_label, fontsize=12)
        axes_lin[count].axis('tight')
        axes_lin[count].set_aspect('auto')
        axes_lin[count].ticklabel_format(style='sci', axis='y', scilimits=(0, 0))
    if len(datasets) > 1:
        axes_lin[count].legend(dataset_names, loc='best')
    if title:
        fig.suptitle(title, fontsize=14)
    plt.tight_layout()
    return fig, axes

###############################################################################


def plot_complex_map_stack(map_stack, num_comps=4, title='Eigenvectors', xlabel='UDVS Step', stdevs=2):
    """
    Plots the provided spectrograms from SVD V vector

    Parameters
    -------------
    map_stack : 3D numpy complex matrices
        Eigenvectors rearranged as - [row, col, component]
    num_comps : int
        Number of components to plot
    title : String
        Title to plot above everything else
    xlabel : String
        Label for x axis
    stdevs : int
        Number of standard deviations to consider for plotting

    Returns
    ---------
    fig, axes
    """
    fig201, axes201 = plt.subplots(2, num_comps, figsize=(4 * num_comps, 8))
    fig201.subplots_adjust(hspace=0.4, wspace=0.4)
    fig201.canvas.set_window_title(title)

    for index in range(num_comps):
        cur_map = np.transpose(map_stack[index, :, :])
        axes = [axes201.flat[index], axes201.flat[index + num_comps]]
        funcs = [np.abs, np.angle]
        labels = ['Amplitude', 'Phase']
        for func, lab, ax in zip(funcs, labels, axes):
            amp_mean = np.mean(func(cur_map))
            amp_std = np.std(func(cur_map))
            ax.imshow(func(cur_map), cmap='inferno',
                      vmin=amp_mean - stdevs * amp_std,
                      vmax=amp_mean + stdevs * amp_std)
            ax.set_title('Eigenvector: %d - %s' % (index + 1, lab))
            ax.set_aspect('auto')
        ax.set_xlabel(xlabel)

    return fig201, axes201


###############################################################################

def plot_complex_loop_stack(loop_stack, x_axis, heading='BE Loops', subtitle='Eigenvector', num_comps=4, x_label=''):
    """
    Plots the provided spectrograms from SVD V vector

    Parameters
    -------------
    loop_stack : 3D numpy complex matrices
        Loops rearranged as - [component, points]
    x_axis : 1D real numpy array
        The vector to plot against
    num_comps : int
        Number of components to plot
    title : String
        Title to plot above everything else
    x_label : String
        Label for x axis
    stdevs : int
        Number of standard deviations to consider for plotting

    Returns
    ---------
    fig, axes
    """
    funcs = [np.abs, np.angle]
    labels = ['Amplitude', 'Phase']

    fig201, axes201 = plt.subplots(len(funcs), num_comps, figsize=(num_comps * 4, 4 * len(funcs)))
    fig201.subplots_adjust(hspace=0.4, wspace=0.4)
    fig201.canvas.set_window_title(heading)

    for index in range(num_comps):
        cur_map = loop_stack[index, :]
        axes = [axes201.flat[index], axes201.flat[index + num_comps]]
        for func, lab, ax in zip(funcs, labels, axes):
            ax.plot(x_axis, func(cur_map))
            ax.set_title('%s: %d - %s' % (subtitle, index + 1, lab))
        ax.set_xlabel(x_label)
    fig201.tight_layout()

    return fig201, axes201

###############################################################################


def plotScree(scree, title='Scree'):
    """
    Plots the scree or scree

    Parameters
    -------------
    scree : 1D real numpy array
        The scree vector from SVD

    Returns
    ---------
    fig, axes
    """
    fig203 = plt.figure(figsize=(6.5, 6))
    axes203 = fig203.add_axes([0.1, 0.1, .8, .8])  # left, bottom, width, height (range 0 to 1)
    axes203.loglog(np.arange(len(scree)) + 1, scree, 'b', marker='*')
    axes203.set_xlabel('Principal Component')
    axes203.set_ylabel('Variance')
    axes203.set_title(title)
    axes203.set_xlim(left=1, right=len(scree))
    axes203.set_ylim(bottom=np.min(scree), top=np.max(scree))
    fig203.canvas.set_window_title("Scree")

    return fig203, axes203


# ###############################################################################


def plot_map_stack(map_stack, num_comps=9, stdevs=2, color_bar_mode=None, evenly_spaced=False,
                   title='Component', heading='Map Stack', fig_mult=(4, 4), pad_mult=(0.1, 0.07), **kwargs):
    """
    Plots the provided stack of maps

    Parameters
    -------------
    map_stack : 3D real numpy array
        structured as [rows, cols, component]
    num_comps : unsigned int
        Number of components to plot
    stdevs : int
        Number of standard deviations to consider for plotting
    color_bar_mode : String, Optional
        Options are None, single or each. Default None
    title : String or list of strings
        The titles for each of the plots.
        If a single string is provided, the plot titles become ['title 01', title 02', ...].
        if a list of strings (equal to the number of components) are provided, these are used instead.
    heading : String
        ###Insert description here### Default 'Map Stack'
    fig_mult : length 2 array_like of uints
        Size multipliers for the figure.  Figure size is calculated as (num_rows*`fig_mult[0]`, num_cols*`fig_mult[1]`).
        Default (4, 4)
    pad_mult : length 2 array_like of floats
        Multipliers for the axis padding between plots in the stack.  Padding is calculated as
        (pad_mult[0]*fig_mult[1], pad_mult[1]*fig_mult[0]) for the width and height padding respectively.
        Default (0.1, 0.07)
    kwargs : dictionary
        Keyword arguments to be passed to either matplotlib.pyplot.figure, mpl_toolkits.axes_grid1.ImageGrid, or
        pycroscopy.vis.plot_utils.plot_map.  See specific function documentation for the relavent options.

    Returns
    ---------
    fig, axes
    """
    num_comps = abs(num_comps)
    num_comps = min(num_comps, map_stack.shape[-1])

    if evenly_spaced:
        chosen_pos = np.linspace(0, map_stack.shape[-1] - 1, num_comps, dtype=int)
    else:
        chosen_pos = np.arange(num_comps, dtype=int)

    if isinstance(title, list):
        if len(title) > num_comps:
            # remove additional titles
            title = title[:num_comps]
        elif len(title) < num_comps:
            # add titles
            title += ['Component' + ' ' + str(x) for x in range(len(title), num_comps)]
    else:
        if not isinstance(title, str):
            title = 'Component'
        title = [title + ' ' + str(x) for x in chosen_pos]

    fig_h, fig_w = fig_mult
    p_rows = int(np.floor(np.sqrt(num_comps)))
    p_cols = int(np.ceil(num_comps / p_rows))
    if p_rows*p_cols < num_comps:
        p_cols += 1

    pad_w, pad_h = pad_mult

    '''
    Set defaults for kwargs to the figure creation and extract any non-default values from current kwargs
    '''
    figkwargs = dict()

    if sys.version_info.major == 3:
        inspec_func = inspect.getfullargspec
    else:
        inspec_func = inspect.getargspec

    for key in inspec_func(plt.figure).args:
        if key in kwargs:
            figkwargs.update({key: kwargs.pop(key)})

    fig202 = plt.figure(figsize=(p_cols * fig_w, p_rows * fig_h), **figkwargs)

    '''
    Set defaults for kwargs to the ImageGrid and extract any non-default values from current kwargs
    '''
    igkwargs = {'cbar_pad': '1%',
                'cbar_size': '5%',
                'cbar_location': 'right',
                'direction': 'row',
                'add_all': True,
                'share_all': False,
                'aspect': True,
                'label_mode': 'L'}
    for key in igkwargs.keys():
        if key in kwargs:
            igkwargs.update({key: kwargs.pop(key)})

    axes202 = ImageGrid(fig202, 111, nrows_ncols=(p_rows, p_cols),
                        cbar_mode=color_bar_mode,
                        axes_pad=(pad_w*fig_w, pad_h*fig_h),
                        **igkwargs)
    fig202.canvas.set_window_title(heading)
    fig202.suptitle(heading, fontsize=16)

    for count, index, subtitle in zip(range(chosen_pos.size), chosen_pos, title):
        im = plot_map(axes202[count],
                      map_stack[:, :, index],
                      stdevs=stdevs, **kwargs)
        axes202[count].set_title(subtitle)
        if color_bar_mode is 'each':
            axes202.cbar_axes[count].colorbar(im)

    if color_bar_mode is 'single':
        axes202.cbar_axes[0].colorbar(im)

    return fig202, axes202


def plot_cluster_h5_group(h5_group, y_spec_label, centroids_together=True):
    """
    Plots the cluster labels and mean response for each cluster

    Parameters
    ----------
    h5_group : h5py.Datagroup object
        H5 group containing the labels and mean response
    y_spec_label : str
        Label to use for Y axis on cluster centroid plot
    centroids_together : Boolean, optional - default = True
        Whether or nor to plot all centroids together on the same plot

    Returns
    -------
    fig : Figure
        Figure containing the plots
    axes : 1D array_like of axes objects
        Axes of the individual plots within `fig`
    """
    # TODO: The quantity and units for the main dataset itself are missing in most cases!
    h5_labels = h5_group['Labels']
    try:
        h5_mean_resp = h5_group['Mean_Response']
    except KeyError:
        # old PySPM format:
        h5_mean_resp = h5_group['Centroids']

    # Reshape the mean response to N dimensions
    mean_response, success = reshape_to_Ndims(h5_mean_resp)

    # unfortunately, we cannot use the above function for the labels
    # However, we will assume that the position values are linked to the labels:
    h5_pos_vals = h5_labels.file[h5_labels.attrs['Position_Values']]
    h5_pos_inds = h5_labels.file[h5_labels.attrs['Position_Indices']]

    # Reshape the labels correctly:
    pos_dims = []
    for col in range(h5_pos_inds.shape[1]):
        pos_dims.append(np.unique(h5_pos_inds[:, col]).size)

    pos_ticks = [h5_pos_vals[:pos_dims[0], 0], h5_pos_vals[slice(0,None,pos_dims[0]), 1]]
    # prepare the axes ticks for the map

    pos_dims.reverse()  # go from slowest to fastest
    pos_dims = tuple(pos_dims)
    label_mat = np.reshape(h5_labels.value, pos_dims)

    # Figure out the correct units and labels for mean response:
    h5_spec_vals = h5_mean_resp.file[h5_mean_resp.attrs['Spectroscopic_Values']]
    x_spec_label = get_formatted_labels(h5_spec_vals)[0]

    # Figure out the correct axes labels for label map:
    pos_labels = get_formatted_labels(h5_pos_vals)
    # TODO: cleaner x and y axes labels instead of 0.0000125 etc.

    if centroids_together:
        return plot_cluster_results_together(label_mat, mean_response, spec_val=np.squeeze(h5_spec_vals[0]),
                                             spec_label=x_spec_label, resp_label=y_spec_label,
                                             pos_labels=pos_labels, pos_ticks=pos_ticks)
    else:
        return plot_cluster_results_separate(label_mat, mean_response, max_centroids=4, x_label=x_spec_label,
                                             spec_val=np.squeeze(h5_spec_vals[0]), y_label=y_spec_label)

###############################################################################


def plot_cluster_results_together(label_mat, mean_response, spec_val=None, cmap=plt.cm.viridis,
                                  spec_label='Spectroscopic Value', resp_label='Response',
                                  pos_labels=('X', 'Y'), pos_ticks=None):
    """
    Plot the cluster labels and mean response for each cluster in separate plots

    Parameters
    ----------
    label_mat : 2D ndarray or h5py.Dataset of ints
        Spatial map of cluster labels structured as [rows, cols]
    mean_response : 2D array or h5py.Dataset
        Mean value of each cluster over all samples 
        arranged as [cluster number, features]
    spec_val :  1D array or h5py.Dataset of floats, optional
        X axis to plot the centroids against
        If no value is specified, the data is plotted against the index
    cmap : plt.cm object or str, optional
        Colormap to use for the labels map and the centroid.
        Advised to pick a map where the centroid plots show clearly.
        Default = matplotlib.pyplot.cm.jet
    spec_label : str, optional
        Label to use for X axis on cluster centroid plot
        Default = 'Spectroscopic Value'
    resp_label : str, optional
        Label to use for Y axis on cluster centroid plot
         Default = 'Response'
    pos_labels : array_like of str, optional
        Labels to use for the X and Y axes on the Label map
        Default = ('X', 'Y')
    pos_ticks : array_like of int

    Returns
    -------
    fig : Figure
        Figure containing the plots
    axes : 1D array_like of axes objects
        Axes of the individual plots within `fig`
    """

    def __plot_centroids(centroids, ax, spec_val, spec_label, y_label, cmap, title=None):
        plot_line_family(ax, spec_val, centroids, label_prefix='Cluster', cmap=cmap)
        ax.set_ylabel(y_label)
        # ax.legend(loc='best')
        if title:
            ax.set_title(title)
            ax.set_xlabel(spec_label)

    if type(spec_val) == type(None):
        spec_val = np.arange(mean_response.shape[1])

    if mean_response.dtype in [np.complex64, np.complex128, np.complex]:
        fig = plt.figure(figsize=(12, 8))
        ax_map = plt.subplot2grid((2, 12), (0, 0), colspan=6, rowspan=2)
        ax_amp = plt.subplot2grid((2, 12), (0, 6), colspan=4)
        ax_phase = plt.subplot2grid((2, 12), (1, 6), colspan=4)
        axes = [ax_map, ax_amp, ax_phase]

        __plot_centroids(np.abs(mean_response), ax_amp, spec_val, spec_label,
                        resp_label + ' - Amplitude', cmap, 'Mean Response')
        __plot_centroids(np.angle(mean_response), ax_phase, spec_val, spec_label,
                        resp_label + ' - Phase', cmap)
        plot_handles, plot_labels = ax_amp.get_legend_handles_labels()

    else:
        fig = plt.figure(figsize=(12, 8))
        ax_map = plt.subplot2grid((1, 12), (0, 0), colspan=6)
        ax_resp = plt.subplot2grid((1, 12), (0, 6), colspan=4)
        axes = [ax_map, ax_resp]
        __plot_centroids(mean_response, ax_resp, spec_val, spec_label,
                        resp_label, cmap, 'Mean Response')
        plot_handles, plot_labels = ax_resp.get_legend_handles_labels()

    fleg = plt.figlegend(plot_handles, plot_labels, loc='center right',
                         borderaxespad=0.0)
    num_clusters = mean_response.shape[0]

    if isinstance(label_mat, h5py.Dataset):
        """
        Reshape label_mat based on linked positions
        """
        pos = label_mat.file[label_mat.attrs['Position_Indices']]
        nx = len(np.unique(pos[:, 0]))
        ny = len(np.unique(pos[:, 1]))
        label_mat = label_mat[()].reshape(nx, ny)

    # im = ax_map.imshow(label_mat, interpolation='none')
    ax_map.set_xlabel(pos_labels[0])
    ax_map.set_ylabel(pos_labels[1])

    if pos_ticks is not None:
        x_ticks = np.linspace(0, label_mat.shape[1] - 1, 5, dtype=np.uint16)
        y_ticks = np.linspace(0, label_mat.shape[0] - 1, 5, dtype=np.uint16)
        ax_map.set_xticks(x_ticks)
        ax_map.set_yticks(y_ticks)
        ax_map.set_xticklabels(pos_ticks[0][x_ticks])
        ax_map.set_yticklabels(pos_ticks[1][y_ticks])

    """divider = make_axes_locatable(ax_map)
    cax = divider.append_axes("right", size="5%", pad=0.05)  # space for colorbar
    fig.colorbar(im, cax=cax, ticks=np.arange(num_clusters),
                 cmap=discrete_cmap(num_clusters, base_cmap=plt.cm.viridis))
    ax_map.axis('tight')"""
    pcol0 = ax_map.pcolor(label_mat, cmap=discrete_cmap(num_clusters, base_cmap=plt.cm.viridis))
    fig.colorbar(pcol0, ax=ax_map, ticks=np.arange(num_clusters))
    ax_map.axis('tight')
    ax_map.set_aspect('auto')
    ax_map.set_title('Cluster Label Map')

    fig.tight_layout()
    fig.canvas.set_window_title('Cluster results')

    return fig, axes

###############################################################################


def plot_cluster_results_separate(label_mat, cluster_centroids, max_centroids=4,
                                  spec_val=None, x_label='Excitation (a.u.)', y_label='Response (a.u.)'):
    """
    Plots the provided labels mat and centroids from clustering

    Parameters
    ----------
    label_mat : 2D int numpy array
                structured as [rows, cols]
    cluster_centroids: 2D real numpy array
                       structured as [cluster,features]
    max_centroids : unsigned int
                    Number of centroids to plot
    spec_val :  array-like
        X axis to plot the centroids against
        If no value is specified, the data is plotted against the index
    x_label : String / unicode
              X label for centroid plots
    y_label : String / unicode
              Y label for centroid plots

    Returns
    -------
    fig
    """

    if max_centroids < 5:

        fig501 = plt.figure(figsize=(20, 10))
        fax1 = plt.subplot2grid((2, 4), (0, 0), colspan=2, rowspan=2)
        fax2 = plt.subplot2grid((2, 4), (0, 2))
        fax3 = plt.subplot2grid((2, 4), (0, 3))
        fax4 = plt.subplot2grid((2, 4), (1, 2))
        fax5 = plt.subplot2grid((2, 4), (1, 3))
        fig501.tight_layout()
        axes_handles = [fax1, fax2, fax3, fax4, fax5]

    else:
        fig501 = plt.figure(figsize=(20, 10))
        # make subplot for cluster map
        fax1 = plt.subplot2grid((3, 6), (0, 0), colspan=3, rowspan=3)  # For cluster map
        fax1.set_xmargin(0.50)
        # make subplot for cluster centers
        fax2 = plt.subplot2grid((3, 6), (0, 3))
        fax3 = plt.subplot2grid((3, 6), (0, 4))
        fax4 = plt.subplot2grid((3, 6), (0, 5))
        fax5 = plt.subplot2grid((3, 6), (1, 3))
        fax6 = plt.subplot2grid((3, 6), (1, 4))
        fax7 = plt.subplot2grid((3, 6), (1, 5))
        fax8 = plt.subplot2grid((3, 6), (2, 3))
        fax9 = plt.subplot2grid((3, 6), (2, 4))
        fax10 = plt.subplot2grid((3, 6), (2, 5))
        fig501.tight_layout()
        axes_handles = [fax1, fax2, fax3, fax4, fax5, fax6, fax7, fax8, fax9, fax10]

    # First plot the labels map:
    pcol0 = fax1.pcolor(label_mat, cmap=discrete_cmap(cluster_centroids.shape[0],
                                                      base_cmap=plt.cm.viridis))
    fig501.colorbar(pcol0, ax=fax1, ticks=np.arange(cluster_centroids.shape[0]))
    fax1.axis('tight')
    fax1.set_aspect('auto')
    fax1.set_title('Cluster Label Map')
    """im = fax1.imshow(label_mat, interpolation='none')
    divider = make_axes_locatable(fax1)
    cax = divider.append_axes("right", size="5%", pad=0.05)  # space for colorbar
    plt.colorbar(im, cax=cax)"""

    if spec_val is None and cluster_centroids.ndim == 2:
        spec_val = np.arange(cluster_centroids.shape[1])

    # Plot results
    for ax, index in zip(axes_handles[1: max_centroids + 1], np.arange(max_centroids)):
        if cluster_centroids.ndim == 2:
            ax.plot(spec_val, cluster_centroids[index, :],
                    color=plt.cm.viridis(int(255 * index / (cluster_centroids.shape[0] - 1))))
            ax.set_xlabel(x_label)
            ax.set_ylabel(y_label)
        elif cluster_centroids.ndim == 3:
            plot_map(ax, cluster_centroids[index])
        ax.set_title('Centroid: %d' % index)

    fig501.subplots_adjust(hspace=0.60, wspace=0.60)
    fig501.tight_layout()

    return fig501


###############################################################################

def plot_cluster_dendrogram(label_mat, e_vals, num_comp, num_cluster, mode='Full', last=None,
                            sort_type='distance', sort_mode=True):
    """
    Creates and plots the dendrograms for the given label_mat and
    eigenvalues

    Parameters
    -------------
    label_mat : 2D real numpy array
        structured as [rows, cols], from KMeans clustering
    e_vals: 3D real numpy array of eigenvalues
        structured as [component, rows, cols]
    num_comp : int
        Number of components used to make eigenvalues
    num_cluster : int
        Number of cluster used to make the label_mat
    mode: str, optional
        How should the dendrograms be created.
        "Full" -- use all clusters when creating the dendrograms
        "Truncated" -- stop showing clusters after 'last'
    last: int, optional - should be provided when using "Truncated"
        How many merged clusters should be shown when using
        "Truncated" mode
    sort_type: {'count', 'distance'}, optional
        What type of sorting should be used when plotting the
        dendrograms.  Options are:
        count - Uses the count_sort from scipy.cluster.hierachy.dendrogram
        distance - Uses the distance_sort from scipy.cluster.hierachy.dendrogram
    sort_mode: {False, True, 'ascending', 'descending'}, optional
        For the chosen sort_type, which mode should be used.
        False - Does no sorting
        'ascending' or True - The child with the minimum of the chosen sort
        parameter is plotted first
        'descending' - The child with the maximum of the chosen sort parameter is
        plotted first

    Returns
    ---------
    fig : matplotlib.pyplot Figure object
        Figure containing the dendrogram
    """
    if mode == 'Truncated' and not last:
        warn('Warning: Truncated dendrograms requested, but no last cluster given.  Reverting to full dendrograms.')
        mode = 'Full'

    if mode == 'Full':
        print('Creating full dendrogram from clusters')
        mode = None
    elif mode == 'Truncated':
        print('Creating truncated dendrogram from clusters.  Will stop at {}.'.format(last))
        mode = 'lastp'
    else:
        raise ValueError('Error: Unknown mode requested for plotting dendrograms. mode={}'.format(mode))

    c_sort = False
    d_sort = False
    if sort_type == 'count':
        c_sort = sort_mode
        if c_sort == 'descending':
            c_sort = 'descendent'
    elif sort_type == 'distance':
        d_sort = sort_mode

    centroid_mat = np.zeros([num_cluster, num_comp])
    for k1 in range(num_cluster):
        [i_x, i_y] = np.where(label_mat == k1)
        u_stack = np.zeros([len(i_x), num_comp])
        for k2 in range(len(i_x)):
            u_stack[k2, :] = np.abs(e_vals[i_x[k2], i_y[k2], :num_comp])

        centroid_mat[k1, :] = np.mean(u_stack, 0)

    # Get the distrance between cluster means
    distance_mat = scipy.spatial.distance.pdist(centroid_mat)

    # get hierachical pairings of clusters
    linkage_pairing = scipy.cluster.hierarchy.linkage(distance_mat, 'weighted')
    linkage_pairing[:, 3] = linkage_pairing[:, 3] / max(linkage_pairing[:, 3])

    fig = plt.figure()
    scipy.cluster.hierarchy.dendrogram(linkage_pairing, p=last, truncate_mode=mode,
                                       count_sort=c_sort, distance_sort=d_sort,
                                       leaf_rotation=90)

    fig.axes[0].set_title('Dendrogram')
    fig.axes[0].set_xlabel('Index or (cluster size)')
    fig.axes[0].set_ylabel('Distance')

    return fig


def plot_1d_spectrum(data_vec, freq, title, figure_path=None):
    """
    Plots the Step averaged BE response

    Parameters
    ------------
    data_vec : 1D numpy array
        Response of one BE pulse
    freq : 1D numpy array
        BE frequency that serves as the X axis of the plot
    title : String
        Plot group name
    figure_path : String / Unicode
        Absolute path of the file to write the figure to

    Returns
    ---------
    fig : Matplotlib.pyplot figure
        Figure handle
    ax : Matplotlib.pyplot axis
        Axis handle
    """
    if len(data_vec) != len(freq):
        warn('plot_1d_spectrum: Incompatible data sizes!!!!')
        print('1D:', data_vec.shape, freq.shape)
        return
    freq *= 1E-3  # to kHz
    fig, ax = plt.subplots(nrows=2, ncols=1, sharex=True)
    ax[0].plot(freq, np.abs(data_vec) * 1E+3)
    ax[0].set_title('Amplitude (mV)')
    ax[1].plot(freq, np.angle(data_vec) * 180 / np.pi)
    ax[1].set_title('Phase (deg)')
    ax[1].set_xlabel('Frequency (kHz)')
    fig.suptitle(title + ': mean UDVS, mean spatial response')
    if figure_path:
        plt.savefig(figure_path, format='png', dpi=300)
    return


###############################################################################

def plot_2d_spectrogram(mean_spectrogram, freq, title, cmap=None, figure_path=None, **kwargs):
    """
    Plots the position averaged spectrogram

    Parameters
    ------------
    mean_spectrogram : 2D numpy complex array
        Means spectrogram arranged as [frequency, UDVS step]
    freq : 1D numpy float array
        BE frequency that serves as the X axis of the plot
    title : String
        Plot group name
    cmap : matplotlib.colors.LinearSegmentedColormap object
        color map. Default = plt.cm.viridis
    figure_path : String / Unicode
        Absolute path of the file to write the figure to

    Returns
    ---------
    fig : Matplotlib.pyplot figure
        Figure handle
    ax : Matplotlib.pyplot axis
        Axis handle
    """
    if mean_spectrogram.shape[1] != len(freq):
        warn('plot_2d_spectrogram: Incompatible data sizes!!!!')
        print('2D:', mean_spectrogram.shape, freq.shape)
        return

    """cmap = kwargs.get('cmap')
    kwargs.pop('cmap')"""
    if cmap is None:  # unpack from kwargs instead
        col_map = plt.cm.viridis  # overriding default

    freq *= 1E-3  # to kHz
    fig, ax = plt.subplots(nrows=2, ncols=1, sharex=True)
    # print(mean_spectrogram.shape)
    # print(freq.shape)
    ax[0].imshow(np.abs(mean_spectrogram), interpolation='nearest', cmap=col_map,
                 extent=[freq[0], freq[-1], mean_spectrogram.shape[0], 0], **kwargs)
    ax[0].set_title('Amplitude')
    # ax[0].set_xticks(freq)
    # ax[0].set_ylabel('UDVS Step')
    ax[0].axis('tight')
    ax[1].imshow(np.angle(mean_spectrogram), interpolation='nearest', cmap=col_map,
                 extent=[freq[0], freq[-1], mean_spectrogram.shape[0], 0], **kwargs)
    ax[1].set_title('Phase')
    ax[1].set_xlabel('Frequency (kHz)')
    # ax[0].set_ylabel('UDVS Step')
    ax[1].axis('tight')
    fig.suptitle(title)
    if figure_path:
        plt.savefig(figure_path, format='png', dpi=300)
    return fig, ax

###############################################################################


def plot_histgrams(p_hist, p_hbins, title, figure_path=None):
    """
    Plots the position averaged spectrogram

    Parameters
    ------------
    p_hist : 2D numpy array
        histogram data arranged as [physical quantity, frequency bin]
    p_hbins : 1D numpy array
        BE frequency that serves as the X axis of the plot
    title : String
        Plot group name
    figure_path : String / Unicode
        Absolute path of the file to write the figure to

    Returns
    ---------
    fig : Matplotlib.pyplot figure
        Figure handle
    """

    base_fig_size = 7
    h_fig = base_fig_size
    w_fig = base_fig_size * 4

    fig = plt.figure(figsize=(w_fig, h_fig))
    fig.suptitle(title)
    iplot = 0

    p_Nx, p_Ny = np.amax(p_hbins, axis=1) + 1

    p_hist = np.reshape(p_hist, (4, p_Ny, p_Nx))

    iplot += 1
    p_plot_title = 'Spectral BEHistogram Amp (log10 of counts)'
    p_plot = fig.add_subplot(1, 4, iplot, title=p_plot_title)
    p_im = p_plot.imshow(np.rot90(np.log10(p_hist[0])), interpolation='nearest')
    p_plot.axis('tight')
    fig.colorbar(p_im, fraction=0.1)

    iplot += 1
    p_plot_title = 'Spectral BEHistogram Phase (log10 of counts)'
    p_plot = fig.add_subplot(1, 4, iplot, title=p_plot_title)
    p_im = p_plot.imshow(np.rot90(np.log10(p_hist[1])), interpolation='nearest')
    p_plot.axis('tight')
    fig.colorbar(p_im, fraction=0.1)

    iplot += 1
    p_plot_title = 'Spectral BEHistogram Real (log10 of counts)'
    p_plot = fig.add_subplot(1, 4, iplot, title=p_plot_title)
    p_im = p_plot.imshow(np.rot90(np.log10(p_hist[2])), interpolation='nearest')
    p_plot.axis('tight')
    fig.colorbar(p_im, fraction=0.1)

    iplot += 1
    p_plot_title = 'Spectral BEHistogram Imag (log10 of counts)'
    p_plot = fig.add_subplot(1, 4, iplot, title=p_plot_title)
    p_im = p_plot.imshow(np.rot90(np.log10(p_hist[3])), interpolation='nearest')
    p_plot.axis('tight')
    fig.colorbar(p_im, fraction=0.1)

    if figure_path:
        plt.savefig(figure_path, format='png')

    return fig


def plot_image_cleaning_results(raw_image, clean_image, stdevs=2, heading='Image Cleaning Results',
                                fig_mult=(4, 4), fig_args={}, **kwargs):
    """
    
    Parameters
    ----------
    raw_image
    clean_image
    stdevs
    color_bar_mode
    fig_mult
    fig_args
    heading

    Returns
    -------

    """
    plot_args = {'cbar_pad': '2.0%', 'cbar_size': '4%', 'hor_axis_pad': 0.115, 'vert_axis_pad': 0.1,
                 'sup_title_size': 26, 'sub_title_size': 22, 'show_x_y_ticks': False, 'show_tick_marks': False,
                 'x_y_tick_font_size': 18, 'cbar_tick_font_size': 18}

    plot_args.update(fig_args)

    fig_h, fig_w = fig_mult
    p_rows = 2
    p_cols = 3

    fig_clean = plt.figure(figsize=(p_cols * fig_w, p_rows * fig_h))
    axes_clean = ImageGrid(fig_clean, 111, nrows_ncols=(p_rows, p_cols), cbar_mode='each',
                           cbar_pad=plot_args['cbar_pad'], cbar_size=plot_args['cbar_size'],
                           axes_pad=(plot_args['hor_axis_pad']*fig_w, plot_args['vert_axis_pad']*fig_h))
    fig_clean.canvas.set_window_title(heading)
    fig_clean.suptitle(heading, fontsize=plot_args['sup_title_size'])

    '''
    Calculate the removed noise and the FFT's of the raw, clean, and noise
    '''
    removed_noise = raw_image - clean_image
    blackman_window_rows = scipy.signal.blackman(clean_image.shape[0])
    blackman_window_cols = scipy.signal.blackman(clean_image.shape[1])

    FFT_raw = np.abs(np.fft.fftshift(
        np.fft.fft2(blackman_window_rows[:, np.newaxis] * raw_image * blackman_window_cols[np.newaxis, :]),
        axes=(0, 1)))
    FFT_clean = np.abs(np.fft.fftshift(
        np.fft.fft2(blackman_window_rows[:, np.newaxis] * clean_image * blackman_window_cols[np.newaxis, :]),
        axes=(0, 1)))
    FFT_noise = np.abs(np.fft.fftshift(
        np.fft.fft2(blackman_window_rows[:, np.newaxis] * removed_noise * blackman_window_cols[np.newaxis, :]),
        axes=(0, 1)))

    '''
    Now find the mean and standard deviation of the images
    '''
    raw_mean = np.mean(raw_image)
    clean_mean = np.mean(clean_image)
    noise_mean = np.mean(removed_noise)

    raw_std = np.std(raw_image)
    clean_std = np.std(clean_image)
    noise_std = np.std(removed_noise)
    fft_clean_std = np.std(FFT_clean)

    '''
    Make lists of everything needed to plot
    '''
    plot_names = ['Original Image', 'Cleaned Image', 'Removed Noise',
                  'FFT Original Image', 'FFT Cleaned Image', 'FFT Removed Noise']
    plot_data = [raw_image, clean_image, removed_noise, FFT_raw, FFT_clean, FFT_noise]
    plot_mins = [raw_mean-stdevs*raw_std, clean_mean-stdevs*clean_std, noise_mean-stdevs*noise_std, 0, 0, 0]
    plot_maxes = [raw_mean+stdevs*raw_std, clean_mean+stdevs*clean_std, noise_mean+stdevs*noise_std,
                  2*stdevs*fft_clean_std, 2*stdevs*fft_clean_std, 2*stdevs*fft_clean_std]

    for count, ax, image, title, min, max in zip(range(6), axes_clean, plot_data, plot_names, plot_mins, plot_maxes):
        im = plot_map(ax, image, stdevs, **kwargs)
        im.set_clim(vmin=min, vmax=max)
        axes_clean[count].set_title(title, fontsize=plot_args['sub_title_size'])
        cbar = axes_clean.cbar_axes[count].colorbar(im)
        cbar.ax.tick_params(labelsize=plot_args['cbar_tick_font_size'])

        if not plot_args['show_x_y_ticks']:
            ax.set_xticklabels([])
            ax.set_yticklabels([])
        if not plot_args['show_tick_marks']:
            ax.get_yaxis().set_visible(False)
            ax.get_xaxis().set_visible(False)

    return fig_clean, axes_clean



# -*- coding: utf-8 -*-
"""
Created on Thursday July 27 2017

@author: Rama Vasudevan
"""

from __future__ import division, print_function, absolute_import, unicode_literals

from os import path, remove, listdir  # File Path formatting
from warnings import warn

import numpy as np  # For array operations
from scipy.io import loadmat
from .translator import Translator
from .utils import generate_dummy_main_parms, build_ind_val_dsets
from ..hdf_utils import getH5DsetRefs, linkRefs
from ..io_hdf5 import ioHDF5  # Now the translator is responsible for writing the data.
from ..microdata import MicroDataGroup, MicroDataset  # building blocks for defining heirarchical storage in the H5 file


class TRKPFMTranslator(Translator):
    """
    Translates  trKPFM datasets from .mat and .dat files to .h5
    """

    def _parse_file_path(self, input_path):
        folder_path, base_name = path.split(input_path)
        base_name = base_name[:-8]

        h5_path = path.join(folder_path, base_name + '.h5')
        if path.exists(h5_path):
            remove(h5_path)

        self.h5_path = h5_path
        # Until a better method is provided....
        self.file_list = list()
        for file in listdir(folder_path):
            if '.dat' in file:
                self.file_list.append(path.join(folder_path, file))


    def _parse_spectrogram_size(self, file_handle):
        """
        
        Parameters
        ----------
        file_handle

        Returns
        -------
        data_length: int, size of the spectrogram
        count: int, number of pixels in dataset +1

        """"""
        """

        f = file_handle

        cont_cond = True
        count = 0

        data_lengths = []

        while cont_cond:
            print(count, f.tell())
            count += 1

            data_length = np.fromfile(f, dtype=np.float32, count=1)

            if data_length > 0:
                data_lengths.append(int(data_length))
                f.seek(int(data_length-1)*4, 1)
            else:
                cont_cond = False

        if len(np.unique(np.array(data_lengths)))>1:
            print("Unequal data lengths! Cannot continue")
        else:
            print("Equal data lengths")

        return data_lengths[0], count


    def translate(self, parm_path):
        """
        The main function that translates the provided file into a .h5 file

        Parameters
        ------------
        parm_path : string / unicode
            Absolute file path of the parameters .mat file.

        Returns
        ----------
        h5_path : string / unicode
            Absolute path of the translated h5 file
        """
        parm_dict, excit_wfm = self._read_parms(parm_path)

        self._parse_file_path(parm_path)

        num_dat_files = len(self.file_list)

        f = open(self.file_list[0], 'rb')
        spectrogram_size, count_vals = self._parse_spectrogram_size(f)
        print("spectrogram size:", spectrogram_size)
        num_pixels = parm_dict['grid_num_rows']*parm_dict['grid_num_cols']
        print('Number of pixels: ', num_pixels)
        print('Count Values: ', count_vals)
        if (num_pixels+1) != count_vals:
            print("Data size does not match number of pixels expected. Cannot continue")

        # Now start creating datasets and populating:
        ds_spec_inds, ds_spec_vals = build_ind_val_dsets([excit_wfm.size], is_spectral=True,
                                                         labels=['Bias'], units=['V'], verbose=False)

        ds_spec_vals.data = np.atleast_2d(excit_wfm)  # The data generated above varies linearly. Override.

        ds_pos_ind, ds_pos_val = build_ind_val_dsets([parm_dict['grid_num_rows'],parm_dict['grid_num_cols']],
                                                     is_spectral=False,
                                                     labels=['X', 'Y'], units=['au', 'au'], verbose=False)

        ds_raw_data = MicroDataset('Raw_Data', data=[],
                                   maxshape=(ds_pos_ind.shape[0], spectrogram_size-5),
                                   dtype=np.complex64, chunking=(1, spectrogram_size-5), compression='gzip')
        ds_raw_data.attrs['quantity'] = ['Complex']

        aux_ds_names = ['Position_Indices', 'Position_Values',
                        'Spectroscopic_Indices', 'Spectroscopic_Values']

        num_ai_chans = np.int(num_dat_files/2) # Division by 2 due to real/imaginary

        # technically should change the date, etc.
        spm_data = MicroDataGroup('')
        global_parms = generate_dummy_main_parms()
        global_parms['data_type'] = 'trKPFM'
        global_parms['translator'] = 'trKPFM'
        spm_data.attrs = global_parms
        meas_grp = MicroDataGroup('Measurement_000')
        meas_grp.attrs = parm_dict
        spm_data.addChildren([meas_grp])

        hdf = ioHDF5(self.h5_path)
        # spm_data.showTree()
        hdf.writeData(spm_data, print_log=False)

        self.raw_datasets = list()

        for chan_index in range(num_ai_chans):
            chan_grp = MicroDataGroup('{:s}{:03d}'.format('Channel_', chan_index), '/Measurement_000/')

            if chan_index == 0:

                chan_grp.attrs = {'Harmonic': 1}
            else:
                chan_grp.attrs = {'Harmonic': 2}

            chan_grp.addChildren([ds_pos_ind, ds_pos_val, ds_spec_inds, ds_spec_vals,
                                      ds_raw_data])
            h5_refs = hdf.writeData(chan_grp, print_log=False)
            h5_raw = getH5DsetRefs(['Raw_Data'], h5_refs)[0]
            linkRefs(h5_raw, getH5DsetRefs(aux_ds_names, h5_refs))
            self.raw_datasets.append(h5_raw)
            self.raw_datasets.append(h5_raw)

        # Now that the N channels have been made, populate them with the actual data....
        self._read_data(parm_dict, parm_path, spectrogram_size)

        hdf.close()
        return self.h5_path

    def _read_data(self, parm_dict, parm_path, data_length):
        """
        Reads raw data and populates the h5 datasets

        Parameters
        ----------
        parm_dict : Dictionary
            dictionary containing parameters for this data
        folder_path : string / unicode
            Absolute path of folder containing the data
        """
        #Determine number of pixels
        num_pixels = parm_dict['grid_num_rows'] * parm_dict['grid_num_cols']

        #The four files in TRKPFM are for real and imaginary parts for 1st, 2nd harmonic
        #Create a list of [True,False,True,False] so files can be written to
        #the appropraite channel

        real_imag = np.zeros(shape=(len(self.file_list),1))
        real_imag[::2] = 1
        real_cond =[]
        for entry in real_imag:
            if entry>0:
                real_cond.append(True)
            else:
                real_cond.append(False)

        #Scan through all the .dat files available
        for ifile, file_path in enumerate(self.file_list):
            f = open(file_path, 'rb')
            results_p = self.read_file(data_length, f)
            spectrogram_matrix = np.array(results_p[:])
            b_axis = spectrogram_matrix.shape[2]
            c_axis = spectrogram_matrix.shape[1]
            dall = np.transpose(spectrogram_matrix, (0, 2, 1)).reshape(num_pixels * c_axis, b_axis)
            _, ia, ic = np.unique(dall, axis=0, return_index=True, return_inverse=True)
            reprowind = np.setdiff1d(ic, ia)

            if len(reprowind>0):
                dall[reprowind, :] = np.nan

            #Write to the datasets
            h5_main = self.raw_datasets[ifile]

            if real_cond[ifile]:
                h5_main[:, :] = dall.reshape(h5_main.shape) + 1j*0
            else:
                h5_main[:, :] += 0 + 1j*dall.reshape(h5_main.shape)

            h5_main.file.flush()

    def read_file(self, data_length, f):
        start_point = 0
        count=0
        count_vals=[]
        f.seek(start_point * 4, 0)
        cont_cond=True
        results_p=[]
        while cont_cond:
            count_vals.append(count)
            count += 1

            data_vec = np.fromfile(f, dtype=np.float32, count=int(data_length))
            data_vec1 = data_vec[5:int(data_length)]

            if len(data_vec) > 1:

                s1 = data_vec[3]
                s2 = data_vec[4]
                #print('Data_mat and s1,s2:', data_vec1.shape, s1, s2)
                data_mat1 = data_vec1.reshape(int(s2), int(s1)).T
                results_p.append(data_mat1)

            else:
                cont_cond = False

        f.close()
        return results_p

    @staticmethod
    def _read_parms(parm_path):
        """
        Copies experimental parameters from the .mat file to a dictionary

        Parameters
        ----------
        parm_path : string / unicode
            Absolute path of the parameters file

        Returns
        -------
        parm_dict : dictionary
            Dictionary containing all relevant parameters
        excit_wfm : 1d numpy float array
            Excitation waveform containing the full DC amplitude vector
        """

        h5_f = loadmat(parm_path)
        parm_dict = dict()

        parm_dict['IO_samp_rate_[Hz]'] = np.uint32(h5_f['IO_rate'][0][0])
        parm_dict['IO_time'] = np.uint32(h5_f['IO_time'][0][0])

        excit_wfm = np.float32(np.squeeze(h5_f['dc_amp_vec']))

        parm_dict['grid_num_rows'] = np.int(h5_f['num_rows'][0][0])
        parm_dict['grid_num_cols'] = np.int(h5_f['num_cols'][0][0])

        return parm_dict, excit_wfm



# -*- coding: utf-8 -*-
"""
Created on Tue Nov  3 15:24:12 2015

@author: Chris R. Smith
"""

from __future__ import division, print_function, absolute_import, unicode_literals
from warnings import warn
import h5py
import os
import numpy as np
from .translator import Translator
from ..hdf_utils import get_attr, link_as_main, checkAndLinkAncillary, findH5group, findDataset, \
    create_spec_inds_from_vals


class LabViewH5Patcher(Translator):
    """
    Patches the hdf5 files from the LabView V3 data aquisition software to meet the
    standards of the Pycroscopy data format.

    """

    def __init__(self):
        super(LabViewH5Patcher, self).__init__()

    def _parse_file_path(self, input_path):
        pass

    def _read_data(self):
        pass

    def translate(self, h5_path, force_patch=False):
        """
        Add the needed references and attributes to the h5 file that are not created by the
        LabView data aquisition program.

        Parameters
        ----------
        h5_path : str
            path to the h5 file
        force_patch : bool, optional
            Should the check to see if the file has already been patched be ignored.
            Default False.

        Returns
        -------
        h5_file : h5py.File
            patched hdf5 file

        """
        # Open the file and check if a patch is needed
        h5_file = h5py.File(os.path.abspath(h5_path), 'r+')
        if h5_file.attrs.get('translator') is not None and not force_patch:
            print('File is already Pycroscopy ready.')
            return h5_file

        '''
        Get the list of all Raw_Data Datasets
        Loop over the list and update the needed attributes
        '''
        raw_list = findDataset(h5_file, 'Raw_Data')
        for _, h5_raw in raw_list:
            # Grab the channel and measurement group of the data to check some needed attributes
            h5_chan = h5_raw.parent
            try:
                c_type = get_attr(h5_chan, 'channel_type')

            except KeyError:
                warn_str = "'channel_type' was not found as an attribute of {}.\n".format(h5_chan.name)
                warn_str +="If this is BEPS or BELine data from the LabView aquisition software, " + \
                           "please run the following piece of code.  Afterwards, run this function again.\n" + \
                           "CODE: " \
                           "hdf.file['{}'].attrs['channel_type'] = 'BE'".format(h5_chan.name)
                warn(warn_str)
                return h5_file

            except:
                raise

            if c_type != 'BE':
                continue

            h5_meas = h5_chan.parent
            h5_meas.attrs['num_UDVS_steps'] = h5_meas.attrs['num_steps']

            # Get the object handles for the Indices and Values datasets
            h5_pos_inds = h5_chan['Position_Indices']
            h5_pos_vals = h5_chan['Position_Values']
            h5_spec_inds = h5_chan['Spectroscopic_Indices']
            h5_spec_vals = h5_chan['Spectroscopic_Values']

            # Make sure we have correct spectroscopic indices for the given values
            ds_spec_inds = create_spec_inds_from_vals(h5_spec_vals[()])
            if not np.allclose(ds_spec_inds, h5_spec_inds[()]):
                h5_spec_inds[:, :] = ds_spec_inds[:, :]
                h5_file.flush()

            # Get the labels and units for the Spectroscopic datasets
            h5_spec_labels = h5_spec_inds.attrs['labels']
            inds_and_vals = [h5_pos_inds, h5_pos_vals, h5_spec_inds, h5_spec_vals]
            for dset in inds_and_vals:
                spec_labels = dset.attrs['labels']
                try:
                    spec_units = dset.attrs['units']

                    if len(spec_units) != len(spec_labels):
                        raise KeyError

                except KeyError:
                    dset['units'] = ['' for _ in spec_labels]
                except:
                    raise

            for ilabel, label in enumerate(h5_spec_labels):
                label_slice = (slice(ilabel, ilabel + 1), slice(None))
                if label == '':
                    label = 'Step'
                h5_spec_inds.attrs[label] = h5_spec_inds.regionref[label_slice]
                h5_spec_vals.attrs[label] = h5_spec_vals.regionref[label_slice]

            # Link the references to the Indices and Values datasets to the Raw_Data
            link_as_main(h5_raw, h5_pos_inds, h5_pos_vals, h5_spec_inds, h5_spec_vals)

            # Also link the Bin_Frequencies and Bin_Wfm_Type datasets
            h5_freqs = h5_chan['Bin_Frequencies']
            aux_dset_names = ['Bin_Frequencies']
            aux_dset_refs = [h5_freqs.ref]
            checkAndLinkAncillary(h5_raw, aux_dset_names, anc_refs=aux_dset_refs)

            '''
            Get all SHO_Fit groups for the Raw_Data and loop over them
            Get the Guess and Spectroscopic Datasets for each SHO_Fit group
            '''
            sho_list = findH5group(h5_raw, 'SHO_Fit')
            for h5_sho in sho_list:
                h5_sho_guess = h5_sho['Guess']
                h5_sho_spec_inds = h5_sho['Spectroscopic_Indices']
                h5_sho_spec_vals = h5_sho['Spectroscopic_Values']

                # Make sure we have correct spectroscopic indices for the given values
                ds_sho_spec_inds = create_spec_inds_from_vals(h5_sho_spec_inds[()])
                if not np.allclose(ds_sho_spec_inds, h5_sho_spec_inds[()]):
                    h5_sho_spec_inds[:, :] = ds_sho_spec_inds[:, :]

                # Get the labels and units for the Spectroscopic datasets
                h5_sho_spec_labels = get_attr(h5_sho_spec_inds, 'labels')
                link_as_main(h5_sho_guess, h5_pos_inds, h5_pos_vals, h5_sho_spec_inds, h5_sho_spec_vals)
                sho_inds_and_vals = [h5_sho_spec_inds, h5_sho_spec_vals]

                for dset in sho_inds_and_vals:
                    spec_labels = get_attr(dset, 'labels')
                    try:
                        spec_units = get_attr(dset, 'units')

                        if len(spec_units) != len(spec_labels):
                            raise KeyError

                    except KeyError:
                        spec_units = [''.encode('utf-8') for _ in spec_labels]
                        dset.attrs['units'] = spec_units

                    except:
                        raise

                # Make region references in the
                for ilabel, label in enumerate(h5_sho_spec_labels):
                    label_slice = (slice(ilabel, ilabel + 1), slice(None))
                    if label == '':
                        label = 'Step'.encode('utf-8')
                    h5_sho_spec_inds.attrs[label] = h5_sho_spec_inds.regionref[label_slice]
                    h5_sho_spec_vals.attrs[label] = h5_sho_spec_vals.regionref[label_slice]

            h5_file.flush()

        h5_file.attrs['translator'] = 'V3patcher'.encode('utf-8')

        return h5_file



# -*- coding: utf-8 -*-
"""
Created on Mon Mar 28 09:45:08 2016

@author: Suhas Somnath, Chris Smith
"""

from __future__ import division, print_function, absolute_import
import time
from warnings import warn
from multiprocessing import cpu_count
import numpy as np
from sklearn.utils import gen_batches
from sklearn.utils.extmath import randomized_svd

from ..io.hdf_utils import getH5DsetRefs, checkAndLinkAncillary, findH5group, create_empty_dataset, \
    getH5RegRefIndices, createRefFromIndices, checkIfMain, calc_chunks, copy_main_attributes, copyAttributes
from ..io.io_hdf5 import ioHDF5
from ..io.io_utils import check_dtype, transformToTargetType, getAvailableMem
from ..io.microdata import MicroDataset, MicroDataGroup

def doSVD(h5_main, num_comps=None):
    """
    Does SVD on the provided dataset and writes the result. File is not closed

    Parameters
    ----------
    h5_main : h5py.Dataset reference
        Reference to the dataset on which SVD will be performed
    num_comps : Unsigned integer (Optional)
        Number of principal components of interest

    Returns
    -------
    h5_pca : h5py.Datagroup reference
        Reference to the group containing the PCA results
    """

    if not checkIfMain(h5_main):
        warn('Dataset does not meet requirements for performing PCA.')
        return

    dset_name = h5_main.name.split('/')[-1]

    t1 = time.time()

    '''
    Calculate the size of the main data in memory and compare to max_mem
    We use the minimum of the actual dtype's itemsize and float32 since we
    don't want to read it in yet and do the proper type conversions.
    '''
    func, is_complex, is_compound, n_features, n_samples, type_mult = check_dtype(h5_main)

    if num_comps is None:
        num_comps = min(n_samples, n_features)
    else:
        num_comps = min(n_samples, n_features, num_comps)

    '''
    Check if a number of compnents has been set and ensure that the number is less than
    the minimum axis length of the data.  If both conditions are met, use fsvd.  If not
    use the regular svd.

    C.Smith -- We might need to put a lower limit on num_comps in the future.  I don't
               know enough about svd to be sure.
    '''
    print('Performing SVD decomposition')

    U, S, V = randomized_svd(func(h5_main), num_comps, n_iter=3)

    svd_type = 'sklearn-randomized'

    print('SVD took {} seconds.  Writing results to file.'.format(round(time.time() - t1, 2)))

    '''
    Create datasets for V and S, deleting original arrays afterward to save
    memory.
    '''
    ds_S = MicroDataset('S', data=np.float32(S))
    ds_S.attrs['labels'] = {'Principal Component': [slice(0, None)]}
    ds_S.attrs['units'] = ''
    ds_inds = MicroDataset('Component_Indices', data=np.uint32(np.arange(len(S))))
    ds_inds.attrs['labels'] = {'Principal Component': [slice(0, None)]}
    ds_inds.attrs['units'] = ''
    del S

    u_chunks = calc_chunks(U.shape, np.float32(0).itemsize)
    ds_U = MicroDataset('U', data=np.float32(U), chunking=u_chunks)
    del U

    # if is_complex:
    #     # Put the real and imaginary sections together to make complex V
    #     V = V[:, :int(0.5 * V.shape[1])] + 1j * V[:, int(0.5 * V.shape[1]):]
    #     v_chunks = calc_chunks(V.shape, h5_main.dtype.itemsize)
    #     ds_V = MicroDataset('V', data=np.complex64(V), chunking=v_chunks)
    # elif is_compound:
    #     V2 = np.empty([V.shape[0], h5_main.shape[1]], dtype=h5_main.dtype)
    #     for iname, name in enumerate(h5_main.dtype.names):
    #         istart = iname * V2.shape[1]
    #         iend = (iname + 1) * V2.shape[1]
    #         V2[name] = V[:, istart:iend]
    #     v_chunks = calc_chunks(V2.shape, h5_main.dtype.itemsize)
    #     ds_V = MicroDataset('V', data=V2, chunking=v_chunks)
    #     del V2
    # else:
    #     v_chunks = calc_chunks(V.shape, h5_main.dtype.itemsize)
    #     ds_V = MicroDataset('V', data=np.float32(V), chunking=v_chunks)
    V = transformToTargetType(V, h5_main.dtype)
    v_chunks = calc_chunks(V.shape, h5_main.dtype.itemsize)
    ds_V = MicroDataset('V', data=V, chunking=v_chunks)
    del V

    '''
    Create the Group to hold the results and add the existing datasets as
    children
    '''
    grp_name = dset_name + '-SVD_'
    svd_grp = MicroDataGroup(grp_name, h5_main.parent.name[1:])
    svd_grp.addChildren([ds_V, ds_S, ds_U, ds_inds])

    '''
    Write the attributes to the group
    '''
    svd_grp.attrs['num_components'] = num_comps
    svd_grp.attrs['svd_method'] = svd_type

    '''
    Write the data and retrieve the HDF5 objects then delete the Microdatasets
    '''
    hdf = ioHDF5(h5_main.file)
    h5_svd_refs = hdf.writeData(svd_grp)

    h5_U = getH5DsetRefs(['U'], h5_svd_refs)[0]
    h5_S = getH5DsetRefs(['S'], h5_svd_refs)[0]
    h5_V = getH5DsetRefs(['V'], h5_svd_refs)[0]
    h5_svd_inds = getH5DsetRefs(['Component_Indices'], h5_svd_refs)[0]
    h5_svd_grp = h5_S.parent

    # copy attributes
    copy_main_attributes(h5_main, h5_V)
    h5_V.attrs['units'] = np.array(['a. u.'], dtype='S')

    del ds_S, ds_V, ds_U, svd_grp

    # Will attempt to see if there is anything linked to this dataset.
    # Since I was meticulous about the translators that I wrote, I know I will find something here
    checkAndLinkAncillary(h5_U,
                          ['Position_Indices', 'Position_Values'],
                          h5_main=h5_main)

    checkAndLinkAncillary(h5_V,
                          ['Position_Indices', 'Position_Values'],
                          anc_refs=[h5_svd_inds, h5_S])

    checkAndLinkAncillary(h5_U,
                          ['Spectroscopic_Indices', 'Spectroscopic_Values'],
                          anc_refs=[h5_svd_inds, h5_S])

    checkAndLinkAncillary(h5_V,
                          ['Spectroscopic_Indices', 'Spectroscopic_Values'],
                          h5_main=h5_main)

    '''
    Check h5_main for plot group references.
    Copy them into V if they exist
    '''
    for key in h5_main.attrs.keys():
        if '_Plot_Group' not in key:
            continue

        ref_inds = getH5RegRefIndices(h5_main.attrs[key], h5_main, return_method='corners')
        ref_inds = ref_inds.reshape([-1, 2, 2])
        ref_inds[:, 1, 0] = h5_V.shape[0] - 1

        svd_ref = createRefFromIndices(h5_V, ref_inds)

        h5_V.attrs[key] = svd_ref

    return h5_svd_grp


###############################################################################

def simplifiedKPCA(kpca, source_data):
    """
    Performs kernel PCA on the provided dataset and returns the familiar
    eigenvector, eigenvalue, and scree matrices.

    Note that the positions in the eigenvalues may need to be transposed

    Parameters
    ----------
    kpca : KernelPCA object
        configured Kernel PCA object ready to perform analysis
    source_data : 2D numpy array
        Data arranged as [iteration, features] example - [position, time]

    Returns
    -------
    eigenvalues : 2D numpy array
        Eigenvalues in the original space arranged as [component,iteration]
    scree : 1D numpy array
        S component
    eigenvector : 2D numpy array
        Eigenvectors in the original space arranged as [component,features]

    """
    X_kpca = kpca.fit(source_data.T)
    eigenvectors = X_kpca.alphas_.T
    eigenvalues = X_kpca.fit_transform(source_data)
    # kpca_explained_variance = np.var(kpca.fit_transform(source_data), axis=0)
    # information_content = kpca_explained_variance / np.sum(kpca_explained_variance)
    scree = kpca.lambdas_
    return eigenvalues, scree, eigenvectors

def rebuild_svd(h5_main, components=None, cores=None, max_RAM_mb=1024):
    """
    Rebuild the Image from the SVD results on the windows
    Optionally, only use components less than n_comp.

    Parameters
    ----------
    h5_main : hdf5 Dataset
        dataset which SVD was performed on
    components : {int, iterable of int, slice} optional
        Defines which components to keep
        Default - None, all components kept

        Input Types
        integer : Components less than the input will be kept
        length 2 iterable of integers : Integers define start and stop of component slice to retain
        other iterable of integers or slice : Selection of component indices to retain
    cores : int, optional
        How many cores should be used to rebuild
        Default - None, all but 2 cores will be used, min 1
    max_RAM_mb : int, optional
        Maximum ammount of memory to use when rebuilding, in Mb.
        Default - 1024Mb

    Returns
    -------
    rebuilt_data : HDF5 Dataset
        the rebuilt dataset

    """

    hdf = ioHDF5(h5_main.file)
    comp_slice = _get_component_slice(components)
    dset_name = h5_main.name.split('/')[-1]

    # Ensuring that at least one core is available for use / 2 cores are available for other use
    max_cores = max(1, cpu_count() - 2)
    #         print('max_cores',max_cores)
    if cores is not None:
        cores = min(round(abs(cores)), max_cores)
    else:
        cores = max_cores

    max_memory = min(max_RAM_mb*1024**2, 0.75*getAvailableMem())
    if cores != 1:
        max_memory = int(max_memory/2)

    '''
    Get the handles for the SVD results
    '''
    try:
        h5_svd = findH5group(h5_main, 'SVD')[-1]

        h5_S = h5_svd['S']
        h5_U = h5_svd['U']
        h5_V = h5_svd['V']

    except KeyError:
        warnstring = 'SVD Results for {dset} were not found.'.format(dset=dset_name)
        warn(warnstring)
        return
    except:
        raise

    func, is_complex, is_compound, n_features, n_samples, type_mult = check_dtype(h5_V)

    '''
    Calculate the size of a single batch that will fit in the available memory
    '''
    n_comps = h5_S[comp_slice].size
    mem_per_pix = (h5_U.dtype.itemsize+h5_V.dtype.itemsize*h5_V.shape[1])*n_comps
    fixed_mem = h5_main.size * h5_main.dtype.itemsize

    if cores is None:
        free_mem = max_memory - fixed_mem
    else:
        free_mem = max_memory * 2 - fixed_mem

    batch_size = int(round(float(free_mem) / mem_per_pix))
    batch_slices = gen_batches(h5_U.shape[0], batch_size)

    print('Reconstructing in batches of {} positions.'.format(batch_size))
    print('Batchs should be {} Mb each.'.format(mem_per_pix*batch_size/1024.0**2))

    '''
    Loop over all batches.
    '''
    ds_V = np.dot(np.diag(h5_S[comp_slice]), func(h5_V[comp_slice, :]))
    rebuild = np.zeros((h5_main.shape[0], ds_V.shape[1]))
    for ibatch, batch in enumerate(batch_slices):
        rebuild[batch, :] += np.dot(h5_U[batch, comp_slice], ds_V)

    rebuild = transformToTargetType(rebuild, h5_V.dtype)

    print('Completed reconstruction of data from SVD results.  Writing to file.')
    '''
    Create the Group and dataset to hold the rebuild data
    '''
    rebuilt_grp = MicroDataGroup('Rebuilt_Data_', h5_svd.name[1:])

    ds_rebuilt = MicroDataset('Rebuilt_Data', rebuild,
                              chunking=h5_main.chunks,
                              compression=h5_main.compression)
    rebuilt_grp.addChildren([ds_rebuilt])

    if isinstance(comp_slice, slice):
        rebuilt_grp.attrs['components_used'] = '{}-{}'.format(comp_slice.start, comp_slice.stop)
    else:
        rebuilt_grp.attrs['components_used'] = components

    h5_refs = hdf.writeData(rebuilt_grp)

    h5_rebuilt = getH5DsetRefs(['Rebuilt_Data'], h5_refs)[0]
    copyAttributes(h5_main, h5_rebuilt, skip_refs=False)

    hdf.flush()

    print('Done writing reconstructed data to file.')

    return h5_rebuilt


def _get_component_slice(components):
    """
    Check the components object to determine how to use it to slice the dataset

    Parameters
    ----------
    components : {int, iterable of ints, slice, or None}
        Input Options
        integer: Components less than the input will be kept
        length 2 iterable of integers: Integers define start and stop of component slice to retain
        other iterable of integers or slice: Selection of component indices to retain
        None: All components will be used
    Returns
    -------
    comp_slice : slice or numpy array of uints
        Slice or array specifying which components should be kept

    """

    comp_slice = slice(None)

    if isinstance(components, int):
        # Component is integer
        comp_slice = slice(0, components)
    elif hasattr(components, '__iter__') and not isinstance(components, dict):
        # Component is array, list, or tuple
        if len(components) == 2:
            # If only 2 numbers are given, use them as the start and stop of a slice
            comp_slice = slice(int(components[0]), int(components[1]))
        else:
            #Convert components to an unsigned integer array
            comp_slice = np.uint(np.round(components)).tolist()
    elif isinstance(components, slice):
        # Components is already a slice
        comp_slice = components
    elif components is not None:
        raise TypeError('Unsupported component type supplied to clean_and_build.  Allowed types are integer, numpy array, list, tuple, and slice.')

    return comp_slice


"""
Created on 7/17/16 10:08 AM
@author: Suhas Somnath, Numan Laanait, Chris R. Smith
"""

from __future__ import division, print_function, absolute_import, unicode_literals
from warnings import warn
import numpy as np
from .model import Model
from ..io.be_hdf_utils import isReshapable, reshapeToNsteps, reshapeToOneStep
from ..io.hdf_utils import buildReducedSpec, copyRegionRefs, linkRefs, getAuxData, getH5DsetRefs, \
            copyAttributes, get_attr
from ..io.microdata import MicroDataset, MicroDataGroup

'''
Custom dtype for the datasets created during fitting.
'''
# sho32 = np.dtype([('Amplitude [V]', np.float32), ('Frequency [Hz]', np.float32),
#                   ('Quality Factor', np.float32), ('Phase [rad]', np.float32),
#                   ('R2 Criterion', np.float32)])
field_names = ['Amplitude [V]', 'Frequency [Hz]', 'Quality Factor', 'Phase [rad]', 'R2 Criterion']
sho32 = np.dtype({'names': field_names,
                  'formats': [np.float32 for name in field_names]})


class BESHOmodel(Model):
    """
    Analysis of Band excitation spectra with harmonic oscillator responses.
    
    Parameters
    ----------
    h5_main : h5py.Dataset instance
        The dataset over which the analysis will be performed. This dataset should be linked to the spectroscopic
        indices and values, and position indices and values datasets.
    variables : list(string), Default ['Frequency']
        Lists of attributes that h5_main should possess so that it may be analyzed by Model.
    parallel : bool, optional
        Should the parallel implementation of the fitting be used.  Default True.

    Returns
    -------
    None
    
    """

    def __init__(self, h5_main, variables=['Frequency'], parallel=True):
        super(BESHOmodel, self).__init__(h5_main, variables, parallel)
        self.step_start_inds = None
        self.is_reshapable = True
        self.num_udvs_steps = None
        self.freq_vec = None

    def _create_guess_datasets(self):
        """
        Creates the h5 group, guess dataset, corresponding spectroscopic datasets and also
        links the guess dataset to the spectroscopic datasets.
        """
        # Create all the ancilliary datasets, allocate space.....

        h5_spec_inds = getAuxData(self.h5_main, auxDataName=['Spectroscopic_Indices'])[0]
        h5_spec_vals = getAuxData(self.h5_main, auxDataName=['Spectroscopic_Values'])[0]

        self.step_start_inds = np.where(h5_spec_inds[0] == 0)[0]
        self.num_udvs_steps = len(self.step_start_inds)

        # find the frequency vector and hold in memory
        self._get_frequency_vector()

        self.is_reshapable = isReshapable(self.h5_main, self.step_start_inds)

        ds_guess = MicroDataset('Guess', data=[],
                                maxshape=(self.h5_main.shape[0], self.num_udvs_steps),
                                chunking=(1, self.num_udvs_steps), dtype=sho32)

        not_freq = get_attr(h5_spec_inds, 'labels') != 'Frequency'

        ds_sho_inds, ds_sho_vals = buildReducedSpec(h5_spec_inds, h5_spec_vals, not_freq, self.step_start_inds)

        dset_name = self.h5_main.name.split('/')[-1]
        sho_grp = MicroDataGroup('-'.join([dset_name,
                                           'SHO_Fit_']),
                                 self.h5_main.parent.name[1:])
        sho_grp.addChildren([ds_guess,
                             ds_sho_inds,
                             ds_sho_vals])
        sho_grp.attrs['SHO_guess_method'] = "pycroscopy BESHO"

        h5_sho_grp_refs = self.hdf.writeData(sho_grp)

        self.h5_guess = getH5DsetRefs(['Guess'], h5_sho_grp_refs)[0]
        h5_sho_inds = getH5DsetRefs(['Spectroscopic_Indices'],
                                    h5_sho_grp_refs)[0]
        h5_sho_vals = getH5DsetRefs(['Spectroscopic_Values'],
                                    h5_sho_grp_refs)[0]

        # Reference linking before actual fitting
        linkRefs(self.h5_guess, [h5_sho_inds, h5_sho_vals])
        # Linking ancillary position datasets:
        aux_dsets = getAuxData(self.h5_main, auxDataName=['Position_Indices', 'Position_Values'])
        linkRefs(self.h5_guess, aux_dsets)

        copyRegionRefs(self.h5_main, self.h5_guess)

    def _create_fit_datasets(self):
        """
        Creates the HDF5 fit dataset. pycroscopy requires that the h5 group, guess dataset,
        corresponding spectroscopic and position datasets be created and populated at this point.
        This function will create the HDF5 dataset for the fit and link it to same ancillary datasets as the guess.
        The fit dataset will NOT be populated here but will instead be populated using the __setData function
        """

        if self.h5_guess is None:
            warn('Need to guess before fitting!')
            return

        if self.step_start_inds is None:
            h5_spec_inds = getAuxData(self.h5_main, auxDataName=['Spectroscopic_Indices'])[0]
            self.step_start_inds = np.where(h5_spec_inds[0] == 0)[0]

        if self.num_udvs_steps is None:
            self.num_udvs_steps = len(self.step_start_inds)

        if self.freq_vec is None:
            self._get_frequency_vector()

        h5_sho_grp = self.h5_guess.parent

        sho_grp = MicroDataGroup(h5_sho_grp.name.split('/')[-1],
                                 h5_sho_grp.parent.name[1:])

        # dataset size is same as guess size
        ds_result = MicroDataset('Fit', data=[], maxshape=(self.h5_guess.shape[0], self.h5_guess.shape[1]),
                                 chunking=self.h5_guess.chunks, dtype=sho32)
        sho_grp.addChildren([ds_result])
        sho_grp.attrs['SHO_fit_method'] = "pycroscopy BESHO"

        h5_sho_grp_refs = self.hdf.writeData(sho_grp)

        self.h5_fit = getH5DsetRefs(['Fit'], h5_sho_grp_refs)[0]

        '''
        Copy attributes of the fit guess
        '''
        copyAttributes(self.h5_guess, self.h5_fit, skip_refs=False)

    def _get_frequency_vector(self):
        """
        Reads the frequency vector from the Spectroscopic_Values dataset.  
        This assumes that the data is reshape-able.
        
        """
        h5_spec_vals = getAuxData(self.h5_main, auxDataName=['Spectroscopic_Values'])[0]
        freq_dim = np.argwhere(get_attr(h5_spec_vals, 'labels') == 'Frequency').squeeze()

        if len(self.step_start_inds) == 1:  # BE-Line
            end_ind = h5_spec_vals.shape[1]
        else:  # BEPS
            end_ind = self.step_start_inds[1]

        self.freq_vec = h5_spec_vals[freq_dim, self.step_start_inds[0]:end_ind]

    def _get_data_chunk(self, verbose=False):
        """
        Returns the next chunk of data for the guess or the fit

        Parameters
        ----------
        verbose : Boolean (optional. default = False)
            Whether or not to print debug statements
        """

        """
        # The model class should take care of all the basic reading
        super(BESHOmodel, self)._get_data_chunk(verbose=verbose)
        """

        if self._start_pos < self.h5_main.shape[0]:
            self._end_pos = int(min(self.h5_main.shape[0], self._start_pos + self._max_pos_per_read))
            self.data = self.h5_main[self._start_pos:self._end_pos, :]
            if verbose:
                print('Reading pixels {} to {} of {}'.format(self._start_pos, self._end_pos, self.h5_main.shape[0]))

            # Now update the start position
            self._start_pos = self._end_pos
        else:
            if verbose:
                print('Finished reading all data!')
            self.data = None

        # At this point the self.data object is the raw data that needs to be reshaped to a single UDVS step:
        if self.data is not None:
            if verbose:
                print('Got raw data of shape {} from super'.format(self.data.shape))
            self.data = reshapeToOneStep(self.data, self.num_udvs_steps)
            if verbose:
                print('Reshaped raw data to shape {}'.format(self.data.shape))

    def _get_guess_chunk(self):
        """
        Returns the next chunk of the guess dataset corresponding to the main dataset.
        
        """

        # if self.data is None:
        #     self._end_pos = int(min(self.h5_main.shape[0], self._start_pos + self._max_pos_per_read))
        #     self.guess = self.h5_guess[self._start_pos:self._end_pos, :]
        # else:
        #     self.guess = self.h5_guess[self._start_pos:self._end_pos, :]
        self._end_pos = int(min(self.h5_main.shape[0], self._start_pos + self._max_pos_per_read))
        self.guess = self.h5_guess[self._start_pos:self._end_pos, :]
        # At this point the self.data object is the raw data that needs to be reshaped to a single UDVS step:
        self.guess = reshapeToOneStep(self.guess, self.num_udvs_steps)
        # don't keep the R^2.
        self.guess = np.hstack([self.guess[name] for name in self.guess.dtype.names if name != 'R2 Criterion'])
        # bear in mind that this self.guess is a compound dataset.

    def _set_results(self, is_guess=False, verbose=False):
        """
        Writes the provided chunk of data into the guess or fit datasets. 
        This method is responsible for any and all book-keeping.

        Parameters
        ---------
        is_guess : Boolean
            Flag that differentiates the guess from the fit
        verbose : Boolean (optional. default = False)
            Whether or not to print debug statements
        
        """
        if is_guess:
            # prepare to reshape:
            self.guess = np.transpose(np.atleast_2d(self.guess))
            if verbose:
                print('Prepared guess of shape {} before reshaping'.format(self.guess.shape))
            self.guess = reshapeToNsteps(self.guess, self.num_udvs_steps)
            if verbose:
                print('Reshaped guess to shape {}'.format(self.guess.shape))
        else:
            self.fit = np.transpose(np.atleast_2d(self.fit))
            self.fit = reshapeToNsteps(self.fit, self.num_udvs_steps)

        # ask super to take care of the rest, which is a standardized operation
        super(BESHOmodel, self)._set_results(is_guess)

    def _set_guess(self, h5_guess):
        """
        Setup to run the fit on an existing guess dataset.  Sets the attributes
        normally defined during do_guess.

        Parameters
        ----------
        h5_guess : h5py.Dataset
            Dataset object containing the guesses
        """
        h5_spec_inds = getAuxData(self.h5_main, auxDataName=['Spectroscopic_Indices'])[0]

        self.step_start_inds = np.where(h5_spec_inds[0] == 0)[0]
        self.num_udvs_steps = len(self.step_start_inds)

        # find the frequency vector and hold in memory
        self._get_frequency_vector()

        self.is_reshapable = isReshapable(self.h5_main, self.step_start_inds)

        self.h5_guess = h5_guess

        if self._parallel:
            self._max_pos_per_read /= 2

    def do_guess(self, max_mem=None, processors=None, strategy='complex_gaussian',
                 options={"peak_widths": np.array([10, 200]), "peak_step": 20}):
        """

        Parameters
        ----------
        max_mem : uint, optional
            Memory in MB to use for computation
            Default None, available memory from psutil.virtual_memory is used
        processors: int
            Number of processors to use during parallel guess
            Default None, output of psutil.cpu_count - 2 is used
        strategy: string
            Default is 'Wavelet_Peaks'.
            Can be one of ['wavelet_peaks', 'relative_maximum', 'gaussian_processes']. For updated list, run GuessMethods.methods
        options: dict
            Default Options for wavelet_peaks{"peaks_widths": np.array([10,200]), "peak_step":20}.
            Dictionary of options passed to strategy. For more info see GuessMethods documentation.

        Returns
        -------
        results : h5py.Dataset object
            Dataset with the SHO guess parameters
            
        """
        if processors is None:
            processors = self._maxCpus
        else:
            processors = min(processors, self._maxCpus)

        if max_mem is not None:
            max_mem = min(max_mem, self._maxMemoryMB)
            self._maxDataChunk = int(max_mem / self._maxCpus)

            # Now calculate the number of positions that can be stored in memory in one go.
            mb_per_position = self.h5_main.dtype.itemsize * self.h5_main.shape[1] / 1024.0 ** 2
            self._max_pos_per_read = int(np.floor(self._maxDataChunk / mb_per_position))

        if self._parallel:
            self._max_pos_per_read = int(self._max_pos_per_read / 2)

        self._create_guess_datasets()
        self._start_pos = 0
        if strategy == 'complex_gaussian':
            freq_vec = self.freq_vec
            options.update({'frequencies': freq_vec})
        super(BESHOmodel, self).do_guess(processors=processors, strategy=strategy, options=options)

        return self.h5_guess

    def do_fit(self, max_mem=None, processors=None, solver_type='least_squares', solver_options={'jac': 'cs'},
               obj_func={'class': 'Fit_Methods', 'obj_func': 'SHO', 'xvals': np.array([])},
               h5_guess=None):
        """
        Fits the dataset to the SHO function

        Parameters
        ----------
        max_mem : uint, optional
            Memory in MB to use for computation
            Default None, available memory from psutil.virtual_memory is used
        processors : int
            Number of processors the user requests.  The minimum of this and self._maxCpus is used.
            Default None
        solver_type : string
            Default is 'Wavelet_Peaks'.
            Can be one of ['wavelet_peaks', 'relative_maximum', 'gaussian_processes']. 
            For updated list, run GuessMethods.methods
        solver_options : dict
            Dictionary of options passed to strategy. For more info see GuessMethods documentation.
            Default {"peaks_widths": np.array([10,200])}}.
        obj_func : dict
            Dictionary defining the class and method containing the function to be fit as well as any 
            additional function parameters.
        h5_guess : h5py.Dataset
            Existing guess to use as input to fit.
            Default None

        Returns
        -------
        results : h5py.Dataset object
            Dataset with the SHO fit parameters
            
        """
        if processors is None:
            processors = self._maxCpus
        else:
            processors = min(processors, self._maxCpus)

        if max_mem is not None:
            max_mem = min(max_mem, self._maxMemoryMB)
            self._maxDataChunk = int(max_mem / self._maxCpus)

            # Now calculate the number of positions that can be stored in memory in one go.
            mb_per_position = self.h5_main.dtype.itemsize * self.h5_main.shape[1] / 1024.0 ** 2
            self._max_pos_per_read = int(np.floor(self._maxDataChunk / mb_per_position))

        if h5_guess is not None or self.h5_guess is None:
            self._set_guess(h5_guess)

        self._create_fit_datasets()
        self._start_pos = 0
        obj_func['xvals'] = self.freq_vec

        super(BESHOmodel, self).do_fit(processors=processors, solver_type=solver_type,
                                       solver_options=solver_options,
                                       obj_func=obj_func)
        return self.h5_fit

    def _reformat_results(self, results, strategy='wavelet_peaks', verbose=False):
        """
        Model specific calculation and or reformatting of the raw guess or fit results

        Parameters
        ----------
        results : array-like
            Results to be formatted for writing
        strategy : str
            The strategy used in the fit.  Determines how the results will be reformatted.
            Default 'wavelet_peaks'
        verbose : bool
            Enables extra print statements if True

        Returns
        -------
        sho_vec : numpy.ndarray
            The reformatted array of parameters.
            
        """
        if verbose:
            print('Strategy to use: {}'.format(strategy))
        # Create an empty array to store the guess parameters
        sho_vec = np.zeros(shape=(len(results)), dtype=sho32)
        if verbose:
            print('Raw results and compound SHO vector of shape {}'.format(len(results)))

        # Extracting and reshaping the remaining parameters for SHO
        if strategy in ['wavelet_peaks', 'relative_maximum', 'absolute_maximum']:
            # wavelet_peaks sometimes finds 0, 1, 2, or more peaks. Need to handle that:
            # peak_inds = np.array([pixel[0] for pixel in results])
            peak_inds = np.zeros(shape=(len(results)), dtype=np.uint32)
            for pix_ind, pixel in enumerate(results):
                if len(pixel) == 1:  # majority of cases - one peak found
                    peak_inds[pix_ind] = pixel[0]
                elif len(pixel) == 0:  # no peak found
                    peak_inds[pix_ind] = int(0.5*self.data.shape[1])  # set to center of band
                else:  # more than one peak found
                    dist = np.abs(np.array(pixel) - int(0.5*self.data.shape[1]))
                    peak_inds[pix_ind] = pixel[np.argmin(dist)]  # set to peak closest to center of band
            if verbose:
                print('Peak positions of shape {}'.format(peak_inds.shape))
            # First get the value (from the raw data) at these positions:
            comp_vals = np.array(
                [self.data[pixel_ind, peak_inds[pixel_ind]] for pixel_ind in np.arange(peak_inds.size)])
            if verbose:
                print('Complex values at peak positions of shape {}'.format(comp_vals.shape))
            sho_vec['Amplitude [V]'] = np.abs(comp_vals)  # Amplitude
            sho_vec['Phase [rad]'] = np.angle(comp_vals)  # Phase in radians
            sho_vec['Frequency [Hz]'] = self.freq_vec[peak_inds]  # Frequency
            sho_vec['Quality Factor'] = np.ones_like(comp_vals) * 10  # Quality factor
            # Add something here for the R^2
            sho_vec['R2 Criterion'] = np.array([self.r_square(self.data, self._sho_func, self.freq_vec, sho_parms) for sho_parms in sho_vec])
        elif strategy in ['complex_gaussian']:
            for iresult, result in enumerate(results):
                sho_vec['Amplitude [V]'][iresult] = result[0]
                sho_vec['Frequency [Hz]'][iresult] = result[1]
                sho_vec['Quality Factor'][iresult] = result[2]
                sho_vec['Phase [rad]'][iresult] = result[3]
                sho_vec['R2 Criterion'][iresult] = result[4]
        elif strategy in ['SHO']:
            for iresult, result in enumerate(results):
                sho_vec['Amplitude [V]'][iresult] = result.x[0]
                sho_vec['Frequency [Hz]'][iresult] = result.x[1]
                sho_vec['Quality Factor'][iresult] = result.x[2]
                sho_vec['Phase [rad]'][iresult] = result.x[3]
                sho_vec['R2 Criterion'][iresult] = 1-result.fun

        return sho_vec
